from selva.ext.templates.mako.settings import MakoTemplateSettings


def modulename_callable():
    pass


def module_writer():
    pass


def preprocessor():
    pass


def include_error_handler():
    pass


class LexerCls:
    pass


cache_args = {
    "type": "file",
    "dir": "/tmp/cache",
}


def test_settings():
    settings = MakoTemplateSettings.model_validate(
        {
            "modulename_callable": f"{modulename_callable.__module__}:{modulename_callable.__qualname__}",
            "module_writer": f"{module_writer.__module__}:{module_writer.__qualname__}",
            "preprocessor": f"{preprocessor.__module__}:{preprocessor.__qualname__}",
            "include_error_handler": f"{include_error_handler.__module__}:{include_error_handler.__qualname__}",
            "lexer_cls": f"{LexerCls.__module__}:{LexerCls.__name__}",
            "cache_args": f"{test_settings.__module__}:cache_args",
        },
    )

    assert settings.modulename_callable is modulename_callable
    assert settings.module_writer is module_writer
    assert settings.preprocessor is preprocessor
    assert settings.include_error_handler is include_error_handler
    assert settings.lexer_cls is LexerCls
    assert settings.cache_args is cache_args



from typing import Annotated

from aiomcache import Client
from asgikit.responses import respond_text

from selva.di import Inject
from selva.web import get


@get
async def index(request, memcached: Annotated[Client, Inject]):
    await memcached.set(b"key", b"value")
    result = (await memcached.get(b"key")).decode("utf-8")

    await respond_text(request.response, result)
    await memcached.delete(b"key")



import pytest

from selva.ext.data.sqlalchemy.settings import SqlAlchemyEngineSettings


@pytest.mark.parametrize(
    "values",
    [
        {"host": "host"},
        {"port": 5432},
        {"database": "database"},
        {"host": "host", "port": 5423},
        {"host": "host", "port": 5432, "database": "database"},
    ],
    ids=[
        "host",
        "port",
        "database",
        "host_port",
        "host_port_database",
    ],
)
def test_sqlalchemy_settings_mutually_exclusive_properties(values: dict):
    with pytest.raises(ValueError):
        SqlAlchemyEngineSettings.model_validate({"url": "url"} | values)



from aiomcache import Client, FlagClient

from selva.configuration.settings import Settings
from selva.di.decorator import service

from .settings import MemcachedSettings


def parse_memcached_address(address: str) -> tuple[str, int]:
    match address.split(":", maxsplit=1):
        case [host]:
            return host, 11211
        case [host, port]:
            port = int(port)
            return host, port


def make_service(name: str):
    @service(name=name if name != "default" else None)
    async def memcached_service(settings: Settings) -> Client:
        memcached_settings = MemcachedSettings.model_validate(
            dict(settings.data.memcached[name])
        )

        if options := memcached_settings.options:
            memcached_options = options.model_dump(exclude_unset=True)
        else:
            memcached_options = {}

        host, port = parse_memcached_address(memcached_settings.address)
        client = FlagClient(host, port, **memcached_options)

        yield client
        await client.close()

    return memcached_service



import inspect
from abc import ABC
from http import HTTPMethod
from typing import Annotated, Any, TypeVar, get_args, get_origin

from asgikit.requests import Body, Request

from selva._util.base_types import get_base_types
from selva._util.maybe_async import maybe_async
from selva.di.container import Container
from selva.di.error import ServiceNotFoundError
from selva.di.inject import Inject
from selva.web.converter.converter import Converter
from selva.web.converter.decorator import register_from_request
from selva.web.converter.error import (
    FromBodyOnWrongHttpMethodError,
    MissingConverterImplError,
    MissingRequestParamExtractorImplError,
)
from selva.web.converter.param_extractor import (
    FromBody,
    FromCookie,
    FromHeader,
    FromPath,
    FromQuery,
    ParamExtractor,
)
from selva.web.exception import HTTPBadRequestException


@register_from_request(FromBody)
class BodyFromRequest(FromBody):
    di: Annotated[Container, Inject]

    async def from_request(
        self,
        request: Request,
        original_type: type,
        parameter_name: str,
        _metadata,
        _optional: bool,
    ) -> Any:
        if request.method not in (HTTPMethod.POST, HTTPMethod.PUT, HTTPMethod.PATCH):
            raise FromBodyOnWrongHttpMethodError(parameter_name)

        if (origin := get_origin(original_type)) is list:
            search_type = get_args(original_type)[0]
            search_types = [
                origin[base_type] for base_type in get_base_types(search_type)
            ]
        elif original_type is dict:
            search_types = [dict]
        else:
            search_types = get_base_types(original_type)

        for base_type in search_types:
            if converter := await self.di.get(
                Converter[Body, base_type], optional=True
            ):
                return await maybe_async(converter.convert(request.body, original_type))

        raise MissingConverterImplError(original_type)


T_EXTRACTOR = TypeVar("T_EXTRACTOR")


class RequestParamFromRequest(ABC):
    di: Annotated[Container, Inject]
    extractor_type: type[T_EXTRACTOR]

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    async def from_request(
        self,
        request: Request,
        original_type: type,
        parameter_name: str,
        metadata: T_EXTRACTOR | type[T_EXTRACTOR],
        optional: bool,
    ):
        parameter_type = metadata if inspect.isclass(metadata) else type(metadata)
        try:
            extractor = await self.di.get(ParamExtractor[parameter_type])
        except ServiceNotFoundError:
            # pylint: disable=raise-missing-from
            raise MissingRequestParamExtractorImplError(parameter_type)

        converter = await self.di.get(Converter[str, original_type])

        if data := extractor.extract(request, parameter_name, metadata):
            return converter.convert(data, original_type)

        if optional:
            return None

        raise HTTPBadRequestException()


@register_from_request(FromPath)
class PathParamFromRequest(RequestParamFromRequest):
    extractor_type = FromPath


@register_from_request(FromQuery)
class QueryParamFromRequest(RequestParamFromRequest):
    extractor_type = FromQuery


@register_from_request(FromHeader)
class HeaderParamFromRequest(RequestParamFromRequest):
    extractor_type = FromHeader


@register_from_request(FromCookie)
class CookieParamFromRequest(RequestParamFromRequest):
    extractor_type = FromCookie



import inspect
from collections import OrderedDict
from collections.abc import Callable
from http import HTTPMethod

import structlog

from selva._util.package_scan import scan_packages
from selva.web.exception import HTTPNotFoundException
from selva.web.routing.decorator import ATTRIBUTE_HANDLER, HandlerInfo
from selva.web.routing.exception import (
    DuplicateRouteError,
    HandlerWithoutDecoratorError,
)
from selva.web.routing.route import Route, RouteMatch

logger = structlog.get_logger()


def _is_handler(arg) -> bool:
    return inspect.iscoroutinefunction(arg) and hasattr(arg, ATTRIBUTE_HANDLER)


class Router:
    def __init__(self):
        self.routes: OrderedDict[str, Route] = OrderedDict()

    def scan(self, *args):
        for item in scan_packages(*args, predicate=_is_handler):
            self.route(item)

    def route(self, handler: Callable):
        handler_info: HandlerInfo = getattr(handler, ATTRIBUTE_HANDLER, None)
        if not handler_info:
            raise HandlerWithoutDecoratorError(handler)

        action_type = handler_info.type
        method: HTTPMethod = action_type.value  # type: ignore
        path = handler_info.path.strip("/")
        route_name = f"{handler.__module__}.{handler.__qualname__}"

        route = Route(method, path, handler, route_name)

        for current_route in self.routes.values():
            if (
                current_route.action != route.action
                and current_route.method == route.method
                and current_route.regex == route.regex
            ):
                raise DuplicateRouteError(route.name, current_route.name)

        self.routes[route_name] = route
        logger.debug(
            "action registered",
            action=f"{handler.__module__}.{handler.__qualname__}",
            method=route.method,
            path=route.path,
        )

    def match(self, method: HTTPMethod | None, path: str) -> RouteMatch | None:
        for route in self.routes.values():
            if (match := route.match(method, path)) is not None:
                return RouteMatch(route, method, path, match)

        return None

    def reverse(self, name: str, **kwargs) -> str:
        if route := self.routes.get(name):
            return route.reverse(**kwargs)

        raise HTTPNotFoundException()



import inspect
import typing
from collections.abc import Callable
from functools import cache
from typing import Annotated

from selva.di.inject import Inject
from selva.web.handler.model import HandlerParams, RequestParam, ServiceParam
from selva.web.routing.exception import HandlerUntypedParametersError


def assert_params_annotated(handler: Callable, *, skip: int):
    """Asserts that parameters past 'skip' are annotated.

    :raises HandlerUntypedParametersError: if parameters are not annotated.
    """
    signature = inspect.signature(handler)
    parameters = list(signature.parameters.values())

    if untyped_params := [
        p.name for p in parameters[skip:] if p.annotation is inspect.Signature.empty
    ]:
        raise HandlerUntypedParametersError(handler, untyped_params)


@cache
def parse_handler_params(handler: Callable, *, skip: int) -> HandlerParams:
    """parse handler parameters from function signature

    returns: mapping of request parameters names to 3-tuple of type,
             metadata and default value
    :raises HandlerUntypedParametersError:
    """

    assert_params_annotated(handler, skip=skip)

    result = HandlerParams([], [])

    signature = inspect.signature(handler, eval_str=True)
    parameters = list(signature.parameters.keys())

    if len(parameters) == 0:
        return result

    if inspect.ismethod(handler):  # function has `self` parameter
        skip += 1

    if len(parameters) <= skip:
        return result

    for name in parameters[skip:]:  # skip request parameter
        type_hint = signature.parameters[name].annotation
        has_default = signature.parameters[name].default is not inspect.Parameter.empty

        if typing.get_origin(type_hint) is not Annotated:
            result.request.append((name, RequestParam(type_hint, None, has_default)))
            continue

        # Annotated is garanteed to have at least 2 args
        param_type, param_meta, *_ = typing.get_args(type_hint)

        # Skip service dependencies
        if param_meta is Inject:
            result.service.append((name, ServiceParam(param_type, None, has_default)))
        elif isinstance(param_meta, Inject):
            result.service.append(
                (name, ServiceParam(param_type, param_meta.name, has_default))
            )
        else:
            result.request.append(
                (name, RequestParam(param_type, param_meta, has_default))
            )

    return result



import asyncio
import traceback
from http import HTTPStatus

import structlog
from asgikit.errors.websocket import WebSocketDisconnectError, WebSocketError
from asgikit.requests import Request
from asgikit.responses import respond_status, respond_text
from asgikit.websockets import WebSocket

from selva._util.import_item import import_item
from selva._util.maybe_async import maybe_async
from selva.configuration.settings import Settings, get_settings
from selva.di.call import call_with_dependencies
from selva.di.container import Container
from selva.ext.error import ExtensionMissingInitFunctionError, ExtensionNotFoundError
from selva.web.exception import HTTPException, HTTPNotFoundException, WebSocketException
from selva.web.exception_handler.discover import find_exception_handlers
from selva.web.handler.call import call_handler
from selva.web.lifecycle.discover import find_background_services, find_startup_hooks
from selva.web.middleware.exception_handler import exception_handler_middleware
from selva.web.routing.router import Router

logger = structlog.get_logger()


def _init_settings(settings: Settings | None) -> Settings:
    if not settings:
        settings = get_settings()

    logging_setup = import_item(settings.logging.setup)
    logging_setup(settings)

    return settings


class Selva:
    """Entrypoint class for a Selva Application

    Will try to automatically import and register a module called "application".
    Other modules and classes can be registered using the "register" method
    """

    def __init__(self, settings: Settings = None, container: Container = None):
        self.settings = _init_settings(settings)

        self.di = Container() if not container else container
        self.di.define(Container, self.di)

        self.di.define(Settings, self.settings)

        self.router = Router()
        self.di.define(Router, self.router)

        self.handler = self._request_handler
        self.exception_handlers = find_exception_handlers(self.settings.application)

        self.startup = find_startup_hooks(self.settings.application)
        self.background_services = find_background_services(self.settings.application)
        self._background_services: set[asyncio.Task] = set()

        self.di.scan(
            self.settings.application,
            "selva.web.converter",
            "selva.web.middleware",
        )
        self.router.scan(self.settings.application)

    async def __call__(self, scope, receive, send):
        match scope["type"]:
            case "http" | "websocket":
                await self._handle_request(scope, receive, send)
            case "lifespan":
                await self._handle_lifespan(scope, receive, send)
            case _:
                raise RuntimeError(f"unknown scope '{scope['type']}'")

    async def _initialize_extensions(self):
        for extension_name in self.settings.extensions:
            try:
                extension_init = import_item(f"{extension_name}:init_extension")
            except ImportError:
                # pylint: disable=raise-missing-from
                raise ExtensionNotFoundError(extension_name)
            except AttributeError:
                # pylint: disable=raise-missing-from
                raise ExtensionMissingInitFunctionError(extension_name)

            await maybe_async(extension_init, self.di, self.settings)

    async def _initialize_middleware(self):
        middleware = self.settings.middleware
        if not middleware:
            middleware_functions = [exception_handler_middleware]
        else:
            middleware_functions = [import_item(name) for name in middleware]
            if exception_handler_middleware in middleware_functions:
                middleware_functions.remove(exception_handler_middleware)

            middleware.append(exception_handler_middleware)

        for factory in reversed(middleware_functions):
            self.handler = await maybe_async(
                factory, self.handler, self.settings, self.di
            )

    async def _lifespan_startup(self):
        for hook in self.startup:
            await call_with_dependencies(self.di, hook)

        for hook in self.background_services:

            async def wrapper():
                try:
                    await call_with_dependencies(self.di, hook)
                except asyncio.CancelledError:
                    pass

            task = asyncio.create_task(wrapper())
            self._background_services.add(task)

            def done_callback(task):
                if err := task.exception():
                    raise err
                self._background_services.discard(task)

            task.add_done_callback(done_callback)

        await self._initialize_extensions()
        await self._initialize_middleware()

    async def _lifespan_shutdown(self):
        for task in self._background_services:
            task.cancel()

        await self.di.run_finalizers()

    async def _handle_lifespan(self, _scope, receive, send):
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                logger.debug("handling lifespan startup")
                try:
                    await self._lifespan_startup()
                    logger.debug("lifespan startup complete")
                    await send({"type": "lifespan.startup.complete"})
                except Exception as err:
                    logger.exception("lifespan startup failed")
                    await send({"type": "lifespan.startup.failed", "message": str(err)})
            elif message["type"] == "lifespan.shutdown":
                logger.debug("handling lifespan shutdown")
                try:
                    await self._lifespan_shutdown()
                    logger.debug("lifespan shutdown complete")
                    await send({"type": "lifespan.shutdown.complete"})
                except Exception as err:
                    logger.debug("lifespan shutdown failed")
                    await send(
                        {"type": "lifespan.shutdown.failed", "message": str(err)}
                    )
                break

    async def _handle_request(self, scope, receive, send):
        request = Request(scope, receive, send)

        try:
            await self.handler(scope, receive, send)
        except (WebSocketDisconnectError, WebSocketError):
            logger.exception("websocket error")
            await request.websocket.close()
        except WebSocketException as err:
            await request.websocket.close(err.code, err.reason)
        except HTTPException as err:
            if websocket := request.websocket:
                logger.exception("websocket request raised HTTPException")
                await websocket.close()
                return

            response = request.response

            if cause := err.__cause__:
                logger.exception(cause)
                stack_trace = "".join(traceback.format_exception(cause))
            else:
                stack_trace = None

            if response.is_started:
                logger.error("response has already started")
                await response.end()
                return

            if response.is_finished:
                logger.error("response is finished")
                return

            if stack_trace:
                response.status = err.status
                await respond_text(response, stack_trace)
            else:
                await respond_status(response, status=err.status)
        except Exception:
            logger.exception("error processing request")

            request.response.status = HTTPStatus.INTERNAL_SERVER_ERROR
            await respond_text(request.response, traceback.format_exc())

    async def _request_handler(self, scope, receive, send):
        request = Request(scope, receive, send)
        path = request.path

        logger.debug(
            "handling request",
            method=str(request.method),
            path=path,
            query=request.query,
        )

        match = self.router.match(request.method, path)

        if not match:
            raise HTTPNotFoundException()

        action = match.route.action
        path_params = match.params
        request["path_params"] = path_params

        await call_handler(self.di, action, request, skip=1)

        response = request.response

        if ws := request.websocket:
            if ws.state != WebSocket.State.CLOSED:
                logger.warning("closing websocket")
                await ws.close()
        elif not response.is_started:
            await respond_status(response, HTTPStatus.INTERNAL_SERVER_ERROR)
        elif not response.is_finished:
            await response.end()



import pytest
from asgikit.requests import Request
from pydantic import BaseModel

from selva.di.container import Container
from selva.web.converter import Form, Json
from selva.web.converter.converter_impl import (
    RequestBodyFormConverter,
    RequestBodyJsonConverter,
    RequestBodyPydanticListConverter,
)
from selva.web.converter.error import (
    FromBodyOnWrongHttpMethodError,
    MissingConverterImplError,
    MissingRequestParamExtractorImplError,
    PathParamNotFoundError,
)
from selva.web.converter.from_request import FromRequest
from selva.web.converter.from_request_impl import (
    BodyFromRequest,
    CookieParamFromRequest,
    HeaderParamFromRequest,
    PathParamFromRequest,
    QueryParamFromRequest,
)
from selva.web.converter.param_converter_impl import StrParamConverter
from selva.web.converter.param_extractor import (
    FromCookie,
    FromHeader,
    FromPath,
    FromQuery,
)
from selva.web.converter.param_extractor_impl import (
    FromCookieExtractor,
    FromHeaderExtractor,
    FromPathExtractor,
    FromQueryExtractor,
)


async def test_body_from_request_wrong_http_method_should_fail(ioc: Container):
    ioc.register(RequestBodyJsonConverter)
    from_request = BodyFromRequest(ioc)

    async def receive():
        return {
            "type": "http.request",
            "body": b'{"field": "value"}',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "GET",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)

    with pytest.raises(FromBodyOnWrongHttpMethodError):
        await from_request.from_request(request, Json, "parameter", None, False)


async def test_body_from_request(ioc: Container):
    ioc.register(RequestBodyJsonConverter)
    from_request = BodyFromRequest(ioc)

    async def receive():
        return {
            "type": "http.request",
            "body": b'{"field": "value"}',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)

    result = await from_request.from_request(request, Json, "parameter", None, False)
    assert isinstance(result, dict)
    assert result["field"] == "value"


async def test_body_list_from_request(ioc: Container):
    ioc.register(RequestBodyPydanticListConverter)
    from_request = BodyFromRequest(ioc)

    class MyModel(BaseModel):
        field: str

    async def receive():
        return {
            "type": "http.request",
            "body": b'[{"field": "value1"}, {"field": "value2"}]',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)

    result = await from_request.from_request(
        request, list[MyModel], "parameter", None, False
    )
    assert isinstance(result, list)
    assert result == [
        MyModel.model_validate({"field": "value1"}),
        MyModel.model_validate({"field": "value2"}),
    ]


async def test_form_from_request(ioc: Container):
    ioc.register(RequestBodyFormConverter)
    from_request = BodyFromRequest(ioc)

    async def receive():
        return {
            "type": "http.request",
            "body": b"field=value",
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/x-www-form-urlencoded")],
    }

    request = Request(scope, receive, None)

    result = await from_request.from_request(request, Form, "parameter", None, False)
    assert isinstance(result, dict)
    assert result["field"] == "value"


async def test_missing_converter_should_fail(ioc: Container):
    from_request = BodyFromRequest(ioc)

    async def receive():
        return {
            "type": "http.request",
            "body": b'{"field": "value"}',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)

    with pytest.raises(MissingConverterImplError):
        await from_request.from_request(request, Json, "parameter", None, False)


async def test_path_param_from_request(ioc: Container):
    ioc.define(Container, ioc)
    ioc.register(StrParamConverter)
    ioc.register(PathParamFromRequest)
    ioc.register(FromPathExtractor)

    from_request = await ioc.get(FromRequest[FromPath])

    scope = {"type": "http", "method": "GET"}
    request = Request(scope, None, None)
    request.attributes["path_params"] = {"param": "value"}

    result = await from_request.from_request(request, str, "param", FromPath, False)
    assert result == "value"


async def test_path_param_from_request_missing_param_should_fail(ioc: Container):
    ioc.define(Container, ioc)
    ioc.register(StrParamConverter)
    ioc.register(PathParamFromRequest)
    ioc.register(FromPathExtractor)

    from_request = await ioc.get(FromRequest[FromPath])

    scope = {"type": "http", "method": "GET"}
    request = Request(scope, None, None)
    request.attributes["path_params"] = {}

    with pytest.raises(PathParamNotFoundError):
        await from_request.from_request(request, str, "param", FromPath, False)


async def test_query_param_from_request(ioc: Container):
    ioc.define(Container, ioc)
    ioc.register(StrParamConverter)
    ioc.register(QueryParamFromRequest)
    ioc.register(FromQueryExtractor)

    from_request = await ioc.get(FromRequest[FromQuery])

    scope = {"type": "http", "method": "GET", "query_string": b"param=value"}
    request = Request(scope, None, None)

    result = await from_request.from_request(request, str, "param", FromQuery, False)
    assert result == "value"


async def test_header_param_from_request(ioc: Container):
    ioc.define(Container, ioc)
    ioc.register(StrParamConverter)
    ioc.register(HeaderParamFromRequest)
    ioc.register(FromHeaderExtractor)

    from_request = await ioc.get(FromRequest[FromHeader])

    scope = {"type": "http", "method": "GET", "headers": [(b"param", b"value")]}
    request = Request(scope, None, None)

    result = await from_request.from_request(request, str, "param", FromHeader, False)
    assert result == "value"


async def test_cookie_param_from_request(ioc: Container):
    ioc.define(Container, ioc)
    ioc.register(StrParamConverter)
    ioc.register(CookieParamFromRequest)
    ioc.register(FromCookieExtractor)

    from_request = await ioc.get(FromRequest[FromCookie])

    scope = {"type": "http", "method": "GET", "headers": [(b"cookie", b"param=value")]}
    request = Request(scope, None, None)

    result = await from_request.from_request(request, str, "param", FromCookie, False)
    assert result == "value"


async def test_missing_param_extractor_should_fail(ioc: Container):
    ioc.define(Container, ioc)
    ioc.register(StrParamConverter)
    ioc.register(PathParamFromRequest)

    from_request = await ioc.get(FromRequest[FromPath])

    scope = {
        "type": "http",
        "method": "GET",
    }
    request = Request(scope, None, None)
    request.attributes["path_params"] = {"param": "value"}

    with pytest.raises(MissingRequestParamExtractorImplError):
        await from_request.from_request(request, str, "param", FromPath, False)



import pytest
import structlog


@pytest.fixture(name="log_output")
def fixture_log_output():
    return structlog.testing.LogCapture()


@pytest.fixture(autouse=True)
def fixture_configure_structlog(log_output):
    structlog.configure(
        processors=[log_output],
    )



default_settings = {
    "application": "application",
    "extensions": [],
    "middleware": [],
    "logging": {
        "setup": "selva.logging:setup",
    },
    "templates": {
        "jinja": {},
        "mako": {},
    },
    "data": {
        "memcached": {},
        "redis": {},
        "sqlalchemy": {},
    },
    "staticfiles": {
        "path": "/static",
        "root": "resources/static",
        "mappings": {},
    },
    "uploadedfiles": {
        "path": "/uploads",
        "root": "resources/uploads",
    },
}



import os
from importlib.util import find_spec

import pytest
from sqlalchemy import make_url

from selva.configuration.settings import _get_settings_nocache
from selva.ext.data.sqlalchemy.service import make_engine_service

POSTGRES_URL = os.getenv("POSTGRES_URL")
SA_DB_URL = make_url(POSTGRES_URL) if POSTGRES_URL else None


async def test_database_url_from_environment_variables(monkeypatch):
    url = "sqlite+aiosqlite:///:memory:"
    monkeypatch.setenv("SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__URL", url)
    settings = _get_settings_nocache()

    async for engine in make_engine_service("default")(settings):
        assert str(engine.url) == url


@pytest.mark.skipif(SA_DB_URL is None, reason="POSTGRES_URL not defined")
@pytest.mark.skipif(find_spec("psycopg") is None, reason="psycopg not present")
async def test_database_url_username_password_from_environment_variables(monkeypatch):
    url = f"postgresql+psycopg://{SA_DB_URL.host}:{SA_DB_URL.port}/{SA_DB_URL.database}"
    username = SA_DB_URL.username
    password = SA_DB_URL.password

    monkeypatch.setenv("SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__URL", url)
    monkeypatch.setenv(
        "SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__USERNAME", username
    )
    monkeypatch.setenv(
        "SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__PASSWORD", password
    )
    settings = _get_settings_nocache()

    async for engine in make_engine_service("default")(settings):
        assert engine.url == SA_DB_URL


@pytest.mark.skipif(SA_DB_URL is None, reason="POSTGRES_URL not defined")
@pytest.mark.skipif(find_spec("psycopg") is None, reason="psycopg not present")
async def test_database_url_components_from_environment_variables(monkeypatch):
    drivername = "postgresql+psycopg"
    host = SA_DB_URL.host
    port = str(SA_DB_URL.port)
    database = SA_DB_URL.database
    username = SA_DB_URL.username
    password = SA_DB_URL.password

    monkeypatch.setenv(
        "SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__DRIVERNAME", drivername
    )
    monkeypatch.setenv("SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__HOST", host)
    monkeypatch.setenv("SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__PORT", port)
    monkeypatch.setenv(
        "SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__DATABASE", database
    )
    monkeypatch.setenv(
        "SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__USERNAME", username
    )
    monkeypatch.setenv(
        "SELVA__DATA__SQLALCHEMY__CONNECTIONS__DEFAULT__PASSWORD", password
    )
    settings = _get_settings_nocache()

    async for engine in make_engine_service("default")(settings):
        assert engine.url == SA_DB_URL



import os
from importlib.util import find_spec

import pytest

from selva.configuration.settings import _get_settings_nocache
from selva.ext.data.memcached.service import make_service

MEMCACHED_ADDR = os.getenv("MEMCACHED_ADDR")


pytestmark = [
    pytest.mark.skipif(MEMCACHED_ADDR is None, reason="MEMCACHED_ADDR not defined"),
    pytest.mark.skipif(find_spec("aiomcache") is None, reason="aiomcache not present"),
]


async def test_address_from_environment_variables(monkeypatch):
    monkeypatch.setenv("SELVA__DATA__MEMCACHED__DEFAULT__ADDRESS", MEMCACHED_ADDR)
    settings = _get_settings_nocache()

    addr = MEMCACHED_ADDR.split(":")
    host = addr[0]
    port = int(addr[1]) if len(addr) > 1 else 11211

    async for memcached in make_service("default")(settings):
        assert memcached._pool._host == host
        assert memcached._pool._port == port



import pytest
from pydantic import BaseModel

from selva._util.pydantic.dotted_path import DottedPath


class Item:
    pass


class MyModel(BaseModel):
    item: DottedPath


def test_dotted_path():
    result = MyModel.model_validate({"item": f"{Item.__module__}:{Item.__qualname__}"})
    assert result.item is Item


def test_invalid_dotted_path():
    with pytest.raises(ValueError):
        MyModel.model_validate({"item": "invalid.dotted:path"})



import pytest
from sqlalchemy import String, select
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.ext.data.sqlalchemy.service import sessionmaker_service


class BaseA(DeclarativeBase):
    pass


class ModelA(BaseA):
    __tablename__ = "model"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(100))


class BaseB(DeclarativeBase):
    pass


class ModelB(BaseB):
    __tablename__ = "model"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(100))


async def test_session_options():
    settings = Settings(
        default_settings
        | {
            "data": {
                "sqlalchemy": {
                    "connections": {
                        "default": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                    },
                    "session": {
                        "options": {
                            "info": {"framework": "selva"},
                        },
                    },
                },
            },
        }
    )

    engine = create_async_engine(settings.data.sqlalchemy.connections.default.url)

    sessionmaker = await sessionmaker_service(settings, {"default": engine})
    async with sessionmaker() as session:
        assert session.info == {"framework": "selva"}


session_info = {
    "framework": "selva",
}


async def test_session_options_info_dotted_path():
    settings = Settings(
        default_settings
        | {
            "data": {
                "sqlalchemy": {
                    "connections": {
                        "default": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                    },
                    "session": {
                        "options": {
                            "info": f"{test_session_options_info_dotted_path.__module__}.session_info",
                        },
                    },
                },
            },
        }
    )

    engine = create_async_engine(settings.data.sqlalchemy.connections.default.url)

    sessionmaker = await sessionmaker_service(settings, {"default": engine})
    async with sessionmaker() as session:
        assert session.info == {"framework": "selva"}


class MySession(AsyncSession):
    pass


async def test_session_class():
    settings = Settings(
        default_settings
        | {
            "data": {
                "sqlalchemy": {
                    "connections": {
                        "default": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                    },
                    "session": {
                        "options": {
                            "class": f"{MySession.__module__}:{MySession.__name__}",
                        },
                    },
                },
            },
        }
    )

    engine = create_async_engine(settings.data.sqlalchemy.connections.default.url)

    sessionmaker = await sessionmaker_service(settings, {"default": engine})
    async with sessionmaker() as session:
        assert session.__class__ == MySession


async def test_binds():
    settings = Settings(
        default_settings
        | {
            "data": {
                "sqlalchemy": {
                    "connections": {
                        "conn_a": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                        "conn_b": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                    },
                    "session": {
                        "binds": {
                            f"{BaseA.__module__}:{BaseA.__qualname__}": "conn_a",
                            f"{BaseB.__module__}:{BaseB.__qualname__}": "conn_b",
                        },
                    },
                },
            },
        }
    )

    engine_a = create_async_engine(settings.data.sqlalchemy.connections.conn_a.url)
    engine_b = create_async_engine(settings.data.sqlalchemy.connections.conn_b.url)

    sessionmaker = await sessionmaker_service(
        settings, {"conn_a": engine_a, "conn_b": engine_b}
    )
    async with sessionmaker() as session:
        assert session.binds[BaseA] == engine_a
        assert session.binds[BaseB] == engine_b


async def test_sessiomaker_without_default_connection():
    settings = Settings(
        default_settings
        | {
            "data": {
                "sqlalchemy": {
                    "connections": {
                        "conn": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                    },
                },
            },
        }
    )

    engine = create_async_engine(settings.data.sqlalchemy.connections.conn.url)

    sessionmaker = await sessionmaker_service(settings, {"conn": engine})
    async with sessionmaker() as session:
        assert session.bind is engine


async def test_binds_model():
    settings = Settings(
        default_settings
        | {
            "data": {
                "sqlalchemy": {
                    "connections": {
                        "conn_a": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                        "conn_b": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                    },
                    "session": {
                        "binds": {
                            f"{BaseA.__module__}:{BaseA.__qualname__}": "conn_a",
                            f"{BaseB.__module__}:{BaseB.__qualname__}": "conn_b",
                        },
                    },
                },
            },
        }
    )

    engine_a = create_async_engine(settings.data.sqlalchemy.connections.conn_a.url)
    async with engine_a.connect() as conn:
        await conn.run_sync(BaseA.metadata.create_all)

    engine_b = create_async_engine(settings.data.sqlalchemy.connections.conn_b.url)
    async with engine_b.connect() as conn:
        await conn.run_sync(BaseB.metadata.create_all)

    sessionmaker = await sessionmaker_service(
        settings, {"conn_a": engine_a, "conn_b": engine_b}
    )
    async with sessionmaker() as session:
        session.add_all(
            [
                ModelA(id=1, name="A"),
                ModelB(id=1, name="B"),
            ]
        )
        await session.commit()

    async with sessionmaker() as session:
        name_a = await session.scalar(select(ModelA.name).where(ModelA.id == 1))
        name_b = await session.scalar(select(ModelB.name).where(ModelB.id == 1))

        assert name_a == "A"
        assert name_b == "B"


async def test_binds_with_invalid_connection_should_fail():
    settings = Settings(
        default_settings
        | {
            "data": {
                "sqlalchemy": {
                    "connections": {
                        "default": {
                            "url": "sqlite+aiosqlite:///:memory:",
                        },
                    },
                    "session": {
                        "binds": {
                            f"{BaseA.__module__}:{BaseA.__qualname__}": "invalid",
                        },
                    },
                },
            },
        }
    )

    with pytest.raises(ValueError, match="No engine with name 'invalid'"):
        await sessionmaker_service(settings, {"default": None})



from typing import Annotated

from selva.di import Inject
from selva.ext.templates.jinja import JinjaTemplate
from selva.web import get


@get("/render")
async def render(request, template: Annotated[JinjaTemplate, Inject]):
    await template.respond(
        request.response,
        "template.html",
        {"variable": "Jinja"},
    )


@get("/stream")
async def stream(request, template: Annotated[JinjaTemplate, Inject]):
    await template.respond(
        request.response,
        "template.html",
        {"variable": "Jinja"},
        stream=True,
    )


@get("/define_content_type")
async def define_content_type(request, template: Annotated[JinjaTemplate, Inject]):
    await template.respond(
        request.response,
        "template.html",
        {"variable": "Jinja"},
        content_type="text/defined",
    )


@get("/override_content_type")
async def override_content_type(request, template: Annotated[JinjaTemplate, Inject]):
    request.response.content_type = "text/plain"

    await template.respond(
        request.response,
        "template.html",
        {"variable": "Jinja"},
        content_type="text/overriden",
    )


@get("/content_type_from_response")
async def content_type_from_response(
    request, template: Annotated[JinjaTemplate, Inject]
):
    request.response.content_type = "text/from_response"

    await template.respond(
        request.response,
        "template.html",
        {"variable": "Jinja"},
    )



import os
from pathlib import Path

import pytest

from selva.configuration.defaults import default_settings
from selva.configuration.settings import (
    Settings,
    SettingsError,
    _get_settings_nocache,
    get_settings,
    get_settings_for_profile,
    merge_recursive,
)


def test_get_settings(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "base")

    result = _get_settings_nocache()
    assert result == default_settings | {
        "prop": "value",
        "list": ["1", "2", "3"],
        "dict": Settings(
            {
                "a": "1",
                "b": "2",
                "c": "3",
            }
        ),
    }


@pytest.mark.parametrize(
    "profile",
    ["dev", "stg", "prd"],
)
def test_get_settings_with_profile(monkeypatch, profile):
    monkeypatch.chdir(Path(__file__).parent / "profiles")
    monkeypatch.setenv("SELVA_PROFILE", profile)

    result = _get_settings_nocache()
    assert result == default_settings | {
        "name": "application",
        "profile": profile,
    }


@pytest.mark.parametrize(
    ["profile_a", "profile_b"],
    [["dev", "stg"], ["stg", "prd"], ["prd", "dev"]],
)
def test_get_settings_with_multiple_profiles(monkeypatch, profile_a, profile_b):
    monkeypatch.chdir(Path(__file__).parent / "multiple_profiles")
    monkeypatch.setenv("SELVA_PROFILE", f"{profile_a}, {profile_b}")

    result = _get_settings_nocache()
    assert result == default_settings | {
        f"profile_{profile_a}": True,
        f"profile_{profile_b}": True,
    }


@pytest.mark.parametrize(
    "profile,expected",
    [
        (None, {"name": "application"}),
        ("dev", {"profile": "dev"}),
        ("stg", {"profile": "stg"}),
        ("prd", {"profile": "prd"}),
    ],
    ids=["None", "dev", "stg", "prd"],
)
def test_get_settings_for_profile(monkeypatch, profile, expected):
    monkeypatch.chdir(Path(__file__).parent / "profiles")

    result = get_settings_for_profile(profile)
    assert result == expected


def test_empty_settings_file(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "empty")

    result = get_settings_for_profile()
    assert result == {}


def test_configure_settings_dir(monkeypatch):
    monkeypatch.setenv(
        "SELVA_SETTINGS_DIR",
        str(Path(__file__).parent / "base" / "configuration"),
    )

    result = _get_settings_nocache()
    assert result == default_settings | {
        "prop": "value",
        "list": ["1", "2", "3"],
        "dict": Settings(
            {
                "a": "1",
                "b": "2",
                "c": "3",
            }
        ),
    }


def test_configure_settings_file(monkeypatch):
    monkeypatch.chdir(str(Path(__file__).parent / "alternate"))
    monkeypatch.setenv(
        "SELVA_SETTINGS_FILE",
        "application.yaml",
    )

    result = _get_settings_nocache()
    assert result == default_settings | {
        "prop": "value",
        "list": ["1", "2", "3"],
        "dict": Settings(
            {
                "a": "1",
                "b": "2",
                "c": "3",
            }
        ),
    }


def test_configure_settings_dir_and_file(monkeypatch):
    monkeypatch.setenv(
        "SELVA_SETTINGS_DIR",
        str(Path(__file__).parent / "alternate" / "configuration"),
    )
    monkeypatch.setenv(
        "SELVA_SETTINGS_FILE",
        "application.yaml",
    )

    result = _get_settings_nocache()
    assert result == default_settings | {
        "prop": "value",
        "list": ["1", "2", "3"],
        "dict": Settings(
            {
                "a": "1",
                "b": "2",
                "c": "3",
            }
        ),
    }


def test_configure_settings_file_with_profile(monkeypatch):
    monkeypatch.chdir(str(Path(__file__).parent / "alternate"))
    monkeypatch.setenv(
        "SELVA_SETTINGS_FILE",
        "application.yaml",
    )

    monkeypatch.setenv("SELVA_PROFILE", "prd")

    result = _get_settings_nocache()
    assert result == default_settings | {
        "profile": "prd",
        "prop": "value",
        "list": ["1", "2", "3"],
        "dict": {
            "a": "1",
            "b": "2",
            "c": "3",
        },
    }


@pytest.mark.parametrize(
    "env",
    ["dev", "stg", "prd"],
)
def test_configure_env_setttings(monkeypatch, env):
    monkeypatch.setenv(
        "SELVA_SETTINGS_DIR",
        str(Path(__file__).parent / "profiles/configuration"),
    )
    monkeypatch.setenv("SELVA_PROFILE", env)

    result = _get_settings_nocache()
    assert result == default_settings | {
        "name": "application",
        "profile": env,
    }


@pytest.mark.parametrize(
    "env",
    ["dev", "stg", "prd"],
)
def test_override_settings(monkeypatch, env):
    monkeypatch.chdir(Path(__file__).parent / "override")

    result = _get_settings_nocache()
    assert result == default_settings | {"value": "base"}

    monkeypatch.setenv("SELVA_PROFILE", env)

    result = _get_settings_nocache()
    assert result == default_settings | {"value": env}


def test_settings_class(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "base")

    settings = _get_settings_nocache()
    assert settings["prop"] == "value"
    assert settings["list"] == ["1", "2", "3"]
    assert settings["dict"] == {
        "a": "1",
        "b": "2",
        "c": "3",
    }


@pytest.mark.parametrize(
    "env",
    ["dev", "stg", "prd"],
)
def test_setttings_class_env(monkeypatch, env):
    monkeypatch.chdir(Path(__file__).parent / "profiles")
    monkeypatch.setenv("SELVA_PROFILE", env)

    settings = _get_settings_nocache()
    assert settings["name"] == "application"
    assert settings["profile"] == env


def test_override_settings_with_env_var(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "base")
    monkeypatch.setenv("SELVA__PROP", "override")

    settings = _get_settings_nocache()

    assert settings.prop == "override"


def test_override_nested_settings_with_env_var(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "base")
    monkeypatch.setenv("SELVA__DICT__A", "override")

    settings = _get_settings_nocache()

    assert settings.dict.a == "override"


def test_non_existent_env_var_should_fail(monkeypatch):
    monkeypatch.chdir(
        Path(__file__).parent / "invalid_configuration" / "non_existent_env_var"
    )

    with pytest.raises(
        ValueError,
        match="DOES_NOT_EXIST environment variable is not defined and does not contain a default value",
    ):
        _get_settings_nocache()


def test_invalid_yaml_should_fail(monkeypatch):
    settings_path = Path(__file__).parent / "invalid_configuration" / "invalid_yaml"
    monkeypatch.chdir(Path(__file__).parent / "invalid_configuration" / "invalid_yaml")

    with pytest.raises(
        SettingsError, match=f"cannot load settings from {settings_path}"
    ):
        get_settings_for_profile()


@pytest.mark.parametrize(
    "settings,extra,expected",
    [
        ({"a": 1}, {"b": 2}, {"a": 1, "b": 2}),
        ({"a": 1}, {"a": 2}, {"a": 2}),
        ({"a": {"a": 1}}, {"a": {"b": 2}}, {"a": {"a": 1, "b": 2}}),
        ({"a": {"a": 1}}, {"a": {"a": 2}}, {"a": {"a": 2}}),
        ({"a": {"a": 1}}, {"a": 1}, {"a": 1}),
        ({"a": 1}, {"a": {"a": 1}}, {"a": {"a": 1}}),
    ],
    ids=[
        "add value",
        "replace value",
        "add nested value",
        "replace nested value",
        "replace dict with value",
        "replace value with dict",
    ],
)
def test_merge_recursive(settings, extra, expected):
    merge_recursive(settings, extra)

    assert settings == expected


def test_settings_item_access():
    settings = Settings({"a": 1})

    assert settings["a"] == 1


def test_settings_nested_item_access():
    settings = Settings({"a": {"b": 1}})

    assert settings["a"]["b"] == 1


def test_settings_attribute_access():
    settings = Settings({"a": 1})

    assert settings.a == 1


def test_settings_nested_attribute_access():
    settings = Settings({"a": {"b": 1}})

    assert settings.a.b == 1


def test_settings_non_existent_attribute_should_fail():
    settings = Settings({})

    with pytest.raises(AttributeError, match="a"):
        settings.a


def test_settings_mixed_item_attribute_access():
    settings = Settings({"a": {"b": 1}})

    assert settings["a"].b == 1
    assert settings.a["b"] == 1


def test_settings_method_access():
    settings = Settings({"a": 1})

    assert settings.get("a") == 1


def test_settings_method_access_default_value():
    settings = Settings({})

    assert settings.get("a") is None
    assert settings.get("a", "b") == "b"


def test_settings_get_nonexistent_item_should_fail():
    settings = Settings({})
    with pytest.raises(KeyError, match="a"):
        settings["a"]


def test_settings_contains_item():
    settings = Settings({"a": 1})
    assert "a" in settings


def test_settings_eq():
    settings1 = Settings({"a": 1})
    settings2 = Settings({"a": 1})
    assert settings1 == settings2


def test_settings_not_eq():
    settings1 = Settings({"a": 1})
    settings2 = Settings({"a": 2})
    assert settings1 != settings2


def test_settings_eq_dict():
    settings = Settings({"a": 1})
    other = {"a": 1}
    assert settings == other


def test_settings_not_eq_dict():
    settings = Settings({"a": 1})
    other = {"a": 2}
    assert settings != other


def test_settings_not_eq_object():
    settings = Settings({})
    assert settings != "settings"


def test_settings_hash():
    data = {"a": 1}
    settings = Settings(data)
    assert hash(settings) == hash(repr(data))


def test_settings_str():
    data = {"a": 1}
    settings = Settings(data)
    assert str(settings) == str(data)


def test_settings_repr():
    data = {"a": 1}
    settings = Settings(data)
    assert repr(settings) == repr(data)


def test_settings_with_env_var(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "env_var")
    monkeypatch.setenv("VAR_NAME", "test")

    settings = _get_settings_nocache()
    assert settings.name == "test"


def test_settings_with_env_var_replaced_in_profile(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent / "env_var")
    monkeypatch.setenv("SELVA_PROFILE", "profile")

    settings = _get_settings_nocache()
    assert settings.name == "profile"


def test_get_settings_cache(monkeypatch):
    monkeypatch.setenv("SELVA__PROP", "test")
    assert "SELVA__PROP" in os.environ
    assert get_settings().prop == "test"

    monkeypatch.setenv("SELVA__ANOTHER_PROP", "test")
    monkeypatch.delenv("SELVA__PROP")
    settings = get_settings()
    assert "SELVA__PROP" not in os.environ
    assert "prop" in settings
    assert "SELVA__ANOTHER_PROP" in os.environ
    assert "another_prop" not in settings



from abc import ABC
from typing import Protocol, TypeVar, runtime_checkable

from asgikit.requests import Request

T = TypeVar("T")


@runtime_checkable
class ParamExtractor(Protocol[T]):
    def extract(self, request: Request, parameter_name: str, metadata: T | type[T]):
        raise NotImplementedError()


class FromBody:
    pass


class FromRequestParam(ABC):
    def __init__(self, name: str = None):
        self.name = name


class FromPath(FromRequestParam):
    pass


class FromQuery(FromRequestParam):
    pass


class FromHeader(FromRequestParam):
    pass


class FromCookie(FromRequestParam):
    pass



from http import HTTPStatus

import pydantic
from asgikit.requests import Body, read_form, read_json
from pydantic import BaseModel as PydanticModel

from selva.web.converter import Form, Json
from selva.web.converter.decorator import register_converter
from selva.web.exception import HTTPBadRequestException, HTTPException


@register_converter(Body, Json)
class RequestBodyJsonConverter:
    async def convert(self, body: Body, _original_type: type) -> dict | list:
        if not body.content_type or "application/json" not in body.content_type:
            raise HTTPException(status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE)

        return await read_json(body)


@register_converter(Body, Form)
class RequestBodyFormConverter:
    async def convert(self, body: Body, _original_type: type) -> dict:
        if (
            not body.content_type
            or "application/x-www-form-urlencoded" not in body.content_type
        ):
            raise HTTPException(status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE)

        return await read_form(body)


@register_converter(Body, PydanticModel)
class RequestBodyPydanticConverter:
    async def convert(
        self,
        body: Body,
        original_type: type[PydanticModel],
    ) -> PydanticModel:
        # TODO: make request body decoding extensible
        if body.content_type and "application/json" in body.content_type:
            data = await read_json(body)
        elif (
            body.content_type
            and "application/x-www-form-urlencoded" in body.content_type
        ):
            data = await read_form(body)
        else:
            raise HTTPException(status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE)

        try:
            return original_type.model_validate(data)
        except pydantic.ValidationError as err:
            raise HTTPBadRequestException() from err


@register_converter(Body, list[PydanticModel])
class RequestBodyPydanticListConverter:
    async def convert(
        self,
        body: Body,
        original_type: type[list[PydanticModel]],
    ) -> list[PydanticModel]:
        if body.content_type and "application/json" in body.content_type:
            data = await read_json(body)
        else:
            raise HTTPException(status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE)

        adapter = pydantic.TypeAdapter(original_type)

        try:
            return adapter.validate_python(data)
        except pydantic.ValidationError as err:
            raise HTTPBadRequestException() from err



from collections.abc import Awaitable
from typing import Protocol, TypeVar, runtime_checkable

__all__ = ("Converter",)


T_FROM = TypeVar("T_FROM")
T_INTO = TypeVar("T_INTO")


@runtime_checkable
class Converter(Protocol[T_FROM, T_INTO]):
    def convert(self, data: T_FROM, original_type: type) -> T_INTO | Awaitable[T_INTO]:
        pass



from decimal import Decimal
from pathlib import PurePath

import pytest

from selva.web.converter.param_converter_impl import (
    BoolParamConverter,
    DecimalParamConverter,
    FloatParamConverter,
    IntParamConverter,
    PurePathParamConverter,
    StrParamConverter,
)
from selva.web.exception import HTTPBadRequestException


async def test_str_request_param():
    converter = StrParamConverter()
    value = "A"
    expected = "A"

    result = converter.convert(value, str)
    assert result == expected


async def test_int_request_param():
    converter = IntParamConverter()
    value = "1"
    expected = 1

    result = converter.convert(value, int)
    assert result == expected


def test_int_request_param_with_invalid_value_should_fail():
    converter = IntParamConverter()
    with pytest.raises(HTTPBadRequestException):
        converter.convert("A", int)


async def test_float_request_param():
    converter = FloatParamConverter()
    value = "1.1"
    expected = 1.1

    result = converter.convert(value, float)
    assert result == expected


def test_float_request_param_with_invalid_value_should_fail():
    converter = FloatParamConverter()
    with pytest.raises(HTTPBadRequestException):
        converter.convert("A", float)


async def test_decimal_request_param():
    converter = DecimalParamConverter()
    value = "1.1"
    expected = Decimal("1.1")

    result = converter.convert(value, Decimal)
    assert result == expected


def test_decimal_request_param_with_invalid_value_should_fail():
    converter = DecimalParamConverter()
    with pytest.raises(HTTPBadRequestException):
        converter.convert("A", Decimal)


@pytest.mark.parametrize(
    "value,expected",
    [
        ("1", True),
        ("True", True),
        ("true", True),
        ("tRuE", True),
        ("0", False),
        ("False", False),
        ("false", False),
        ("fAlSe", False),
    ],
    ids=["1", "True", "true", "tRuE", "0", "False", "false", "fAlSe"],
)
async def test_bool_request_param(value: str, expected: bool):
    converter = BoolParamConverter()

    result = converter.convert(value, bool)
    assert result == expected


async def test_bool_request_param_with_invalid_value_should_fail():
    converter = BoolParamConverter()
    with pytest.raises(HTTPBadRequestException):
        converter.convert("invalid", bool)


async def test_path_request_param():
    converter = PurePathParamConverter()

    value = "/path/subpath"
    expected = PurePath("/path", "subpath")

    result = converter.convert(value, PurePath)
    assert result == expected



import inspect
from collections.abc import Callable
from enum import Enum
from http import HTTPMethod
from typing import NamedTuple

from asgikit.requests import Request

from selva.web.handler.parse import assert_params_annotated
from selva.web.routing.exception import (
    HandlerMissingRequestArgumentError,
    HandlerNotAsyncError,
    HandlerRequestTypeError,
)

ATTRIBUTE_HANDLER = "__selva_web_action__"


class HandlerType(Enum):
    GET = HTTPMethod.GET
    HEAD = HTTPMethod.HEAD
    POST = HTTPMethod.POST
    PUT = HTTPMethod.PUT
    PATCH = HTTPMethod.PATCH
    DELETE = HTTPMethod.DELETE
    OPTIONS = HTTPMethod.OPTIONS
    WEBSOCKET = None

    @property
    def is_websocket(self) -> bool:
        return self is HandlerType.WEBSOCKET


class HandlerInfo(NamedTuple):
    type: HandlerType
    path: str


def route(handler: Callable = None, /, *, method: HTTPMethod | None, path: str | None):
    path = path.strip("/") if path else ""

    def inner(arg: Callable):
        if not inspect.iscoroutinefunction(arg):
            raise HandlerNotAsyncError(handler)

        assert_params_annotated(arg, skip=1)

        params = list(inspect.signature(arg).parameters.values())
        if len(params) < 1:
            raise HandlerMissingRequestArgumentError(handler)

        req_param = params[0].annotation
        if req_param is not inspect.Signature.empty and req_param is not Request:
            raise HandlerRequestTypeError(handler)

        setattr(arg, ATTRIBUTE_HANDLER, HandlerInfo(HandlerType(method), path))
        return arg

    return inner(handler) if handler else inner


def _route(method: HTTPMethod | None, path_or_action: str | Callable):
    if isinstance(path_or_action, str):
        path = path_or_action.strip("/")
        action = None
    else:
        path = ""
        action = path_or_action

    return route(action, method=method, path=path)


def get(path_or_action: str | Callable):
    return _route(HTTPMethod.GET, path_or_action)


def post(path_or_action: str | Callable):
    return _route(HTTPMethod.POST, path_or_action)


def put(path_or_action: str | Callable):
    return _route(HTTPMethod.PUT, path_or_action)


def patch(path_or_action: str | Callable):
    return _route(HTTPMethod.PATCH, path_or_action)


def delete(path_or_action: str | Callable):
    return _route(HTTPMethod.DELETE, path_or_action)


def websocket(path_or_action: str | Callable):
    return _route(None, path_or_action)



import os
from importlib.util import find_spec

import pytest

from selva.configuration import Settings
from selva.configuration.defaults import default_settings
from selva.ext.data.memcached.service import make_service, parse_memcached_address

MEMCACHED_ADDR = os.getenv("MEMCACHED_ADDR")

pytestmark = [
    pytest.mark.skipif(MEMCACHED_ADDR is None, reason="MEMCACHED_ADDR not defined"),
    pytest.mark.skipif(find_spec("aiomcache") is None, reason="aiomcache not present"),
]


async def _test_make_service(settings: Settings):
    service = make_service("default")(settings)
    async for memcached in service:
        await memcached.set(b"test", b"test")
        result = await memcached.get(b"test")
        assert result == b"test"
        await memcached.delete(b"test")


async def test_make_service_with_address():
    settings = Settings(
        default_settings
        | {
            "data": {
                "memcached": {
                    "default": {
                        "address": MEMCACHED_ADDR,
                    },
                },
            },
        }
    )

    await _test_make_service(settings)


async def test_make_service_with_options():
    settings = Settings(
        default_settings
        | {
            "data": {
                "memcached": {
                    "default": {
                        "address": MEMCACHED_ADDR,
                        "options": {
                            "pool_size": 10,
                            "pool_minsize": 1,
                        },
                    },
                },
            },
        }
    )

    service = make_service("default")(settings)
    async for memcached in service:
        assert memcached._pool._maxsize == 10
        assert memcached._pool._minsize == 1


def test_parse_memcached_address_with_port():
    host, port = parse_memcached_address("memcached:11212")
    assert host == "memcached"
    assert port == 11212


def test_parse_memcached_address_without_port():
    host, port = parse_memcached_address("memcached")
    assert host == "memcached"
    assert port == 11211



from httpx import ASGITransport, AsyncClient

from selva.configuration import Settings
from selva.configuration.defaults import default_settings
from selva.web.application import Selva


def my_middleware(app, settings, di):
    async def inner(scope, receive, send):
        async def new_send(event: dict):
            if event["type"] == "http.response.body":
                event["body"] = b"Middleware Ok"
            await send(event)

        await app(scope, receive, new_send)

    return inner


async def test_middleware():
    settings = Settings(
        default_settings
        | {
            "application": f"{__package__}.application",
            "middleware": [f"{__package__}.test_middleware:my_middleware"],
        }
    )
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")
    assert response.text == "Middleware Ok"



from selva.di.decorator import service
from selva.web.converter.converter import Converter
from selva.web.converter.from_request import FromRequest
from selva.web.converter.param_extractor import ParamExtractor


def register_from_request(target: type):
    def inner(cls):
        assert issubclass(cls, FromRequest)
        return service(cls, provides=FromRequest[target])

    return inner


def register_converter(source: type, target: type):
    def inner(cls):
        assert issubclass(cls, Converter)
        return service(cls, provides=Converter[source, target])

    return inner


def register_param_extractor(target: type):
    def inner(cls):
        assert issubclass(cls, ParamExtractor)
        return service(cls, provides=ParamExtractor[target])

    return inner



from typing import Any, NamedTuple


class RequestParam(NamedTuple):
    param_type: type
    param_meta: type | Any | None
    has_default: bool


class ServiceParam(NamedTuple):
    param_type: type
    service_name: str | None
    has_default: bool


class HandlerParams(NamedTuple):
    request: list[tuple[str, RequestParam]]
    service: list[tuple[str, ServiceParam]]



from selva._util import dotenv
from selva.configuration.settings import get_settings
from selva.di.container import Container
from selva.web.application import Selva

dotenv.init()

settings = get_settings()
di_container = Container()

app = Selva(settings, di_container)



import asyncio
import mimetypes
import os
from abc import ABC, abstractmethod
from collections.abc import Callable
from email.utils import formatdate
from pathlib import Path

import structlog
from asgikit.requests import Request
from asgikit.responses import respond_file

from selva.configuration import Settings
from selva.di import Container
from selva.web.exception import HTTPNotFoundException

logger = structlog.get_logger()


class BaseFilesMiddleware(ABC):
    def __init__(self, app: Callable, path: str, root: Path):
        self.app = app
        self.path = path if path.endswith("/") else path + "/"
        self.root = root

    @abstractmethod
    def get_file_to_serve(self, scope: dict) -> str | None:
        pass

    async def __call__(self, scope, receive, send):
        if file_to_serve := self.get_file_to_serve(scope):
            file_to_serve = (self.root / file_to_serve).resolve()
            if not (
                file_to_serve.is_file() and file_to_serve.is_relative_to(self.root)
            ):
                raise HTTPNotFoundException()

            request = Request(scope, receive, send)
            await respond_file(request.response, file_to_serve)
        else:
            await self.app(scope, receive, send)


class UploadedFilesMiddleware(BaseFilesMiddleware):
    def get_file_to_serve(self, scope: dict) -> str | None:
        request_path = scope["path"].lstrip("/")

        if request_path.startswith(self.path):
            return request_path.removeprefix(self.path).lstrip("/")

        return None


class StaticFilesMiddleware(BaseFilesMiddleware):
    def __init__(
        self,
        app,
        path: str,
        root: Path,
        filelist: dict[str, tuple[str, int, str]],
        mappings: dict[str, str],
    ):
        super().__init__(app, path, root)
        self.filelist = filelist
        self.mappings = mappings

    def get_file_to_serve(self, scope: dict) -> str | None:
        request_path = scope["path"].lstrip("/")

        file_to_serve = None
        if file_path := self.mappings.get(request_path):
            file_to_serve = file_path
        elif request_path.startswith(self.path):
            file_path = os.path.join(self.root, request_path.removeprefix(self.path))
            if file_path in self.filelist:
                file_to_serve = file_path

        if file_to_serve:
            file_data = self.filelist[file_to_serve]
            content_type, content_length, last_modified = file_data

            # use Request to set response headers
            request = Request(scope, None, None)
            request.response.content_type = content_type
            request.response.content_length = content_length
            request.response.header("last-modified", last_modified)

            return file_to_serve

        return None

    async def __call__(self, scope, receive, send):
        try:
            await super().__call__(scope, receive, send)
        except HTTPNotFoundException:
            if scope["path"] == "/":
                new_scope = scope | {"path": f"{self.path}/index.html"}
                await super().__call__(new_scope, receive, send)
            else:
                raise


async def static_files_middleware(app, settings: Settings, di: Container):
    settings = settings.staticfiles
    path = settings.path.lstrip("/")
    root = Path(settings.root).resolve().absolute()

    # dict of file path to (content-type, content-length, last-modified)
    filelist: dict[str, tuple[str, int, str]] = {}

    for dirpath, _dirnames, filenames in os.walk(root):
        for filename in filenames:
            file = os.path.join(dirpath, filename)
            stat = await asyncio.to_thread(os.stat, file)

            m_type, _ = mimetypes.guess_type(filename, strict=False)
            m_type = m_type or "application/octet-stream"
            last_modified = formatdate(stat.st_mtime, usegmt=True)

            filelist[os.path.join(dirpath, filename)] = (
                m_type,
                stat.st_size,
                last_modified,
            )

    mappings = {
        name.lstrip("/"): os.path.join(root, value.lstrip("/"))
        for name, value in settings.get("mappings", {}).items()
    }

    if difference := set(mappings.values()).difference(filelist):
        files = ", ".join(difference)
        raise ValueError(f"Static files mappings not found: {files}")

    return StaticFilesMiddleware(app, path, root, filelist, mappings)


def uploaded_files_middleware(app, settings: Settings, di: Container):
    settings = settings.uploadedfiles
    path = settings.path.lstrip("/")
    root = Path(settings.root).resolve()
    return UploadedFilesMiddleware(app, path, root)



import pytest

from selva.di.container import Container
from selva.di.error import NonInjectableTypeError


def test_non_injectable_type_should_fail(ioc: Container):
    obj = ()
    with pytest.raises(NonInjectableTypeError):
        ioc.register(obj)



from typing import Annotated

from asgikit.responses import Response, respond_text
from mako.lookup import TemplateLookup

from selva.configuration import Settings
from selva.di import Inject, service
from selva.ext.templates.mako.settings import MakoTemplateSettings


@service
class MakoTemplate:
    settings: Annotated[Settings, Inject]

    lookup: TemplateLookup = None

    def initialize(self):
        mako_settings = MakoTemplateSettings.model_validate(
            self.settings.templates.mako
        )

        kwargs = mako_settings.model_dump(exclude_none=True)
        self.lookup = TemplateLookup(**kwargs)

    # pylint: disable=too-many-arguments
    async def respond(
        self,
        response: Response,
        template_name: str,
        context: dict,
        *,
        content_type: str | None = None,
    ):
        if content_type:
            response.content_type = content_type
        elif not response.content_type:
            response.content_type = "text/html"

        template = self.lookup.get_template(template_name)
        rendered = template.render(**context)
        await respond_text(response, rendered)

    async def render(self, template_name: str, context: dict) -> str:
        template = self.lookup.get_template(template_name)
        return template.render(**context)

    async def render_str(self, source: str, context: dict) -> str:
        template_hash = str(hash(source))
        if not self.lookup.has_template(template_hash):
            self.lookup.put_string(template_hash, source)

        template = self.lookup.get_template(template_hash)
        return template.render(**context)



from importlib.util import find_spec

from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.ext.templates.mako.service import MakoTemplate

__all__ = ("MakoTemplate",)


async def init_extension(container: Container, _settings: Settings):
    if find_spec("mako") is None:
        raise ModuleNotFoundError("Missing 'mako'. Install 'selva' with 'mako' extra.")

    container.register(MakoTemplate)



from typing import Annotated, Generic, TypeVar

import pytest

from selva.di.container import Container
from selva.di.decorator import service
from selva.di.error import TypeVarInGenericServiceError
from selva.di.inject import Inject

T = TypeVar("T")


class GenericService(Generic[T]):
    pass


@service(provides=GenericService[int])
class Service(GenericService[int]):
    pass


@service
def service_factory() -> GenericService[int]:
    return Service()


@service
class ServiceDepends:
    service: Annotated[GenericService[int], Inject]


@service
def service_factory_depends(dependency: GenericService[int]) -> ServiceDepends:
    instance = ServiceDepends()
    setattr(instance, "service", dependency)
    return instance


async def test_get_generic_service_with_class(ioc: Container):
    ioc.register(Service)

    instance = await ioc.get(GenericService[int])
    assert isinstance(instance, Service)


async def test_get_generic_service_with_factory(ioc: Container):
    ioc.register(service_factory)

    instance = await ioc.get(GenericService[int])
    assert isinstance(instance, Service)


async def test_get_generic_service_with_class_with_dependency(ioc: Container):
    ioc.register(Service)
    ioc.register(ServiceDepends)

    instance = await ioc.get(ServiceDepends)
    assert isinstance(instance, ServiceDepends)
    assert isinstance(instance.service, Service)


async def test_get_generic_service_with_factory_with_dependency(ioc: Container):
    ioc.register(service_factory)
    ioc.register(service_factory_depends)

    instance = await ioc.get(ServiceDepends)
    assert isinstance(instance, ServiceDepends)
    assert isinstance(instance.service, Service)


async def test_type_var_in_generic_service_should_fail(ioc: Container):
    with pytest.raises(TypeVarInGenericServiceError):
        ioc.register(service(Service, provides=GenericService[T]))



import pytest
from structlog.testing import capture_logs

from selva.di.container import Container
from selva.di.decorator import service
from selva.di.error import FactoryMissingReturnTypeError


@service
class Service1:
    pass


@service
def service1_factory() -> Service1:
    yield Service1()
    print("Service1")


class Service2:
    def __init__(self, service1: Service1):
        self.service1 = service1


@service
def service2_factory(service1: Service1) -> Service2:
    yield Service2(service1)
    print("Service2")


class Interface:
    pass


class Implementation(Interface):
    pass


@service(provides=Interface)
def interface_factory() -> Interface:
    yield Implementation()
    print("Interface Implementation")


def test_has_service(ioc: Container):
    ioc.register(service1_factory)
    assert ioc.has(Service1)


async def test_service_with_provided_interface(ioc: Container, capfd):
    ioc.register(interface_factory)

    instance = await ioc.get(Interface)
    assert isinstance(instance, Implementation)

    await ioc.run_finalizers()
    assert capfd.readouterr().out == "Interface Implementation\n"


async def test_inject_singleton(ioc: Container, capfd):
    ioc.register(service1_factory)
    ioc.register(service2_factory)

    instance = await ioc.get(Service2)
    assert isinstance(instance, Service2)
    assert isinstance(instance.service1, Service1)

    other_instance = await ioc.get(Service2)
    assert other_instance is instance
    assert other_instance.service1 is instance.service1

    await ioc.run_finalizers()
    assert capfd.readouterr().out == "Service2\nService1\n"


async def test_interface_implementation(ioc: Container, capfd):
    ioc.register(interface_factory)

    instance = await ioc.get(Interface)
    assert isinstance(instance, Implementation)

    await ioc.run_finalizers()
    assert capfd.readouterr().out == "Interface Implementation\n"


def test_factory_function_without_return_type_should_fail(ioc: Container):
    @service
    async def service_factory():
        pass

    with pytest.raises(FactoryMissingReturnTypeError):
        ioc.register(service_factory)


def test_provides_option_should_log_warning(ioc: Container):
    with capture_logs() as cap_logs:
        ioc.register(interface_factory)

    assert (
        cap_logs[0]["event"] == "option 'provides' on a factory function has no effect"
    )



from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.web.application import Selva
from selva.web.lifecycle.decorator import startup


@startup
def startup():
    print("startup", end="")


async def test_application(capfd):
    settings = Settings(
        default_settings
        | {
            "application": f"{test_application.__module__}",
        }
    )

    app = Selva(settings)

    await app._lifespan_startup()
    assert capfd.readouterr().out == "startup"



import functools
import inspect
from collections.abc import Callable
from typing import Any

from asgikit.requests import Request

from selva._util.maybe_async import maybe_async
from selva.di.container import Container
from selva.web.converter.error import MissingFromRequestImplError
from selva.web.converter.from_request import FromRequest
from selva.web.handler.parse import parse_handler_params


async def call_handler(
    di: Container, handler: Callable, request: Request, *, skip: int
):
    actual_handler = handler
    while isinstance(actual_handler, functools.partial):
        actual_handler = actual_handler.func

    handler_params = parse_handler_params(actual_handler, skip=skip)
    request_params = await params_from_request(di, request, handler_params.request)

    request_services = {
        name: await di.get(service_type, name=service_name, optional=has_default)
        for name, (
            service_type,
            service_name,
            has_default,
        ) in handler_params.service
    }

    await handler(request, **(request_params | request_services))


async def params_from_request(
    di: Container,
    request: Request,
    params: list[tuple[str, Any, bool]],
) -> dict[str, Any]:
    result = {}

    for name, (param_type, param_annotation, has_default) in params:
        if param_annotation:
            if inspect.isclass(param_annotation):
                converter_type = param_annotation
            else:
                converter_type = type(param_annotation)
        else:
            converter_type = param_type

        if from_request_service := await di.get(FromRequest[converter_type]):
            value = await maybe_async(
                from_request_service.from_request,
                request,
                param_type,
                name,
                param_annotation,
                has_default,
            )

            if value:
                result[name] = value
        else:
            raise MissingFromRequestImplError(param_type)

    return result



from typing import Annotated

from asgikit.responses import respond_text
from sqlalchemy import text
from sqlalchemy.ext.asyncio import async_sessionmaker

from selva.di import Inject
from selva.web import get


@get
async def index(request, sessionmaker: Annotated[async_sessionmaker, Inject]):
    async with sessionmaker() as session:
        result = await session.execute(text("select sqlite_version()"))
        version = result.first()[0]

    await respond_text(request.response, version)



import logging
import logging.config
import sys

import structlog

from selva.configuration.settings import Settings


def setup(settings: Settings):
    structlog.configure(
        processors=[structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        logger_factory=structlog.stdlib.LoggerFactory(),
    )

    log_format = settings.logging.get("format")
    if not log_format:
        log_format = "console" if sys.stderr.isatty() else "json"

    if log_format == "json":
        renderer = structlog.processors.JSONRenderer()
    elif log_format == "logfmt":
        renderer = structlog.processors.LogfmtRenderer(bool_as_flag=False)
    elif log_format == "keyvalue":
        renderer = structlog.processors.KeyValueRenderer()
    elif log_format == "console":
        renderer = structlog.dev.ConsoleRenderer()
    else:
        raise ValueError("Unknown log format")

    processors = [
        structlog.stdlib.filter_by_level,
        structlog.contextvars.merge_contextvars,
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.stdlib.add_logger_name,
        structlog.processors.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
    ]

    if not isinstance(renderer, structlog.dev.ConsoleRenderer):
        processors.append(structlog.processors.dict_tracebacks)

    processors.append(renderer)

    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "structlog": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processors": processors,
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "structlog",
            }
        },
        "root": {
            "handlers": ["console"],
            "level": settings.logging.get("root", "WARN").upper(),
        },
        "loggers": {
            module: {"level": level.upper()}
            for module, level in settings.logging.get("level", {}).items()
        },
    }

    logging.config.dictConfig(logging_config)



import pytest

from selva.di.container import Container
from selva.di.error import ServiceWithoutDecoratorError


class ServiceClass:
    pass


def service_factory() -> ServiceClass:
    return ServiceClass()


@pytest.mark.parametrize(
    "injectable", [ServiceClass, service_factory], ids=["class", "function"]
)
def test_register_service_without_decorator_should_fail(injectable, ioc: Container):
    with pytest.raises(ServiceWithoutDecoratorError):
        ioc.register(ServiceClass)



# ruff: noqa: F401

from selva.web.converter import Form, Json
from selva.web.converter.param_extractor import (
    FromBody,
    FromCookie,
    FromHeader,
    FromPath,
    FromQuery,
)
from selva.web.exception_handler.decorator import exception_handler
from selva.web.lifecycle.decorator import background, startup
from selva.web.routing.decorator import delete, get, patch, post, put, websocket



from copy import deepcopy
from pathlib import Path

from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.ext.templates.jinja.service import JinjaTemplate


async def test_render_template():
    path = str(Path(__file__).parent.absolute())
    settings = Settings(default_settings | {"templates": {"jinja": {"paths": [path]}}})

    template = JinjaTemplate(settings)
    template.initialize()
    result = await template.render("template.html", {"variable": "Jinja"})
    assert result == "Jinja"


async def test_render_str():
    settings = Settings(deepcopy(default_settings))
    template = JinjaTemplate(settings)
    template.initialize()
    result = await template.render_str("{{ variable }}", {"variable": "Jinja"})
    assert result == "Jinja"



class ExtensionNotFoundError(Exception):
    def __init__(self, name):
        super().__init__(f"Extension '{name}' not found")


class ExtensionMissingInitFunctionError(Exception):
    def __init__(self, name):
        super().__init__(
            f"Extension '{name}' is missing the 'init_extension()' function"
        )



from http import HTTPStatus

from httpx import ASGITransport, AsyncClient

from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.web.application import Selva


async def test_application():
    settings = Settings(
        default_settings | {"application": f"{__package__}.application"}
    )
    app = Selva(settings)

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")
    assert response.text == "Ok"


async def test_not_found():
    settings = Settings(
        default_settings | {"application": f"{__package__}.application"}
    )
    app = Selva(settings)

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/not-found")
    assert response.status_code == HTTPStatus.NOT_FOUND



import functools
from functools import cache

import structlog
from asgikit.requests import Request

from selva._util.base_types import get_base_types
from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.web.exception_handler.decorator import ExceptionHandlerType
from selva.web.exception_handler.discover import find_exception_handlers
from selva.web.handler.call import call_handler

logger = structlog.get_logger()


def exception_handler_middleware(app, settings: Settings, di: Container):
    exception_handlers = find_exception_handlers(settings.application)
    return ExceptionHandlerMiddleware(app, di, exception_handlers)


class ExceptionHandlerMiddleware:
    def __init__(self, app, di: Container, exception_handlers):
        self.app = app
        self.di = di
        self.exception_handlers = exception_handlers

    async def __call__(self, scope, receive, send):
        try:
            await self.app(scope, receive, send)
        except Exception as err:
            if handler := self._get_exception_handler(err):
                logger.debug(
                    "Handling exception with handler",
                    module=handler.__module__,
                    handler=handler.__qualname__,
                )

                request = Request(scope, receive, send)
                await call_handler(
                    self.di, functools.partial(handler, err), request, skip=2
                )
            else:
                raise

    @cache
    def _get_exception_handler(self, err: BaseException) -> ExceptionHandlerType | None:
        err_type = type(err)
        for base in get_base_types(err_type):
            if handler := self.exception_handlers.get(base):
                return handler

        return None



from typing import Annotated

from asgikit.responses import respond_text
from redis.asyncio import Redis

from selva.di import Inject
from selva.web import get


@get
async def index(request, redis: Annotated[Redis, Inject(name="other")]):
    await redis.set("key", "value")
    result = (await redis.get("key")).decode("utf-8")

    await respond_text(request.response, result)
    await redis.delete("key")



from collections.abc import Callable

from pydantic import BaseModel, ConfigDict

from selva._util.pydantic import DottedPath


class MemcachedOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")

    pool_size: int = None
    pool_minsize: int = None
    get_flat_handler: DottedPath[Callable] = None
    set_flat_handler: DottedPath[Callable] = None
    conn_args: DottedPath[dict] = None


class MemcachedSettings(BaseModel):
    model_config = ConfigDict(extra="forbid")

    address: str
    options: MemcachedOptions = None



from http import HTTPStatus
from typing import Literal

import pytest
from asgikit.requests import Request
from pydantic import BaseModel

from selva.web.converter.converter_impl import (
    RequestBodyFormConverter,
    RequestBodyJsonConverter,
    RequestBodyPydanticConverter,
    RequestBodyPydanticListConverter,
)
from selva.web.exception import HTTPException


async def test_json_from_request():
    async def receive():
        return {
            "type": "http.request",
            "body": b'{"field": "value"}',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyJsonConverter()

    result = await converter.convert(request.body, dict)

    assert type(result) is dict
    assert result["field"] == "value"


async def test_json_from_request_with_wrong_content_type_should_fail():
    async def receive():
        return {
            "type": "http.request",
            "body": b"",
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/x-www-form-urlencoded")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyJsonConverter()

    with pytest.raises(HTTPException) as err:
        await converter.convert(request.body, dict)

    assert err.value.status == HTTPStatus.UNSUPPORTED_MEDIA_TYPE


async def test_form_from_request():
    async def receive():
        return {
            "type": "http.request",
            "body": b"field1=value1&field2=value2",
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/x-www-form-urlencoded")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyFormConverter()

    result = await converter.convert(request.body, dict)

    assert type(result) is dict
    assert result["field1"] == "value1" and result["field2"] == "value2"


async def test_form_from_request_with_wrong_content_type_should_fail():
    async def receive():
        return {
            "type": "http.request",
            "body": b"",
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyFormConverter()

    with pytest.raises(HTTPException) as err:
        await converter.convert(request.body, dict)

    assert err.value.status == HTTPStatus.UNSUPPORTED_MEDIA_TYPE


async def test_pydantic_model_from_json_request():
    class Model(BaseModel):
        field1: str
        field2: int

    async def receive():
        return {
            "type": "http.request",
            "body": b'{"field1": "value", "field2": 1}',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyPydanticConverter()

    result = await converter.convert(request.body, Model)

    assert type(result) is Model
    assert result.field1 == "value"
    assert result.field2 == 1


async def test_pydantic_model_from_form_request():
    class Model(BaseModel):
        field1: str
        field2: int

    async def receive():
        return {
            "type": "http.request",
            "body": b"field1=value&field2=1",
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/x-www-form-urlencoded")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyPydanticConverter()

    result = await converter.convert(request.body, Model)

    assert type(result) is Model
    assert result.field1 == "value"
    assert result.field2 == 1


async def test_pydantic_model_with_wrong_content_type_should_fail():
    class Model(BaseModel):
        field1: str
        field2: int

    async def receive():
        return {
            "type": "http.request",
            "body": b'"value",1',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"text/csv")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyPydanticConverter()

    with pytest.raises(HTTPException) as err:
        await converter.convert(request.body, Model)
    assert err.value.status == HTTPStatus.UNSUPPORTED_MEDIA_TYPE


async def test_pydantic_model_with_invalid_data_should_fail():
    class Model(BaseModel):
        field: Literal["value"]

    async def receive():
        return {
            "type": "http.request",
            "body": b'{"field": "wrong"}',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyPydanticConverter()

    with pytest.raises(HTTPException) as err:
        await converter.convert(request.body, Model)
    assert err.value.status == HTTPStatus.BAD_REQUEST


async def test_pydantic_model_list_from_request():
    class Model(BaseModel):
        field: str

    async def receive():
        return {
            "type": "http.request",
            "body": b'[{"field": "value1"}, {"field": "value2"}]',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyPydanticListConverter()

    result = await converter.convert(request.body, list[Model])

    assert isinstance(result, list)
    assert result[0].field == "value1"
    assert result[1].field == "value2"


async def test_pydantic_model_list_with_wrong_content_type_should_fail():
    class Model(BaseModel):
        field: str

    async def receive():
        return {
            "type": "http.request",
            "body": b"value1\nvalue2",
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"text/csv")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyPydanticListConverter()

    with pytest.raises(HTTPException) as err:
        await converter.convert(request.body, list[Model])

    assert err.value.status == HTTPStatus.UNSUPPORTED_MEDIA_TYPE


async def test_pydantic_model_list_with_invalid_data_should_fail():
    class Model(BaseModel):
        field: Literal["value"]

    async def receive():
        return {
            "type": "http.request",
            "body": b'[{"field": "wrong"}]',
            "more_body": False,
        }

    scope = {
        "type": "http",
        "method": "POST",
        "headers": [(b"content-type", b"application/json")],
    }

    request = Request(scope, receive, None)
    converter = RequestBodyPydanticListConverter()

    with pytest.raises(HTTPException) as err:
        await converter.convert(request.body, list[Model])
    assert err.value.status == HTTPStatus.BAD_REQUEST



from selva.configuration.environment import (
    parse_settings_from_env,
    replace_variables_recursive,
    replace_variables_with_env,
)


def test_parse_settings():
    source = {
        "SELVA__ENV1": "1",
        "SELVA__PATH1__ENV2": "2",
        "SELVA__PATH1__PATH2__SUB_ENV3": "3",
        "SELVA__PATH1__PATH2__SUB_ENV4": "4",
        "SELVA__": "None",
        "NOT_SELVA_ENV": "None",
    }

    result = parse_settings_from_env(source)

    expected = {
        "env1": "1",
        "path1": {
            "env2": "2",
            "path2": {
                "sub_env3": "3",
                "sub_env4": "4",
            },
        },
    }
    assert result == expected


def test_replace_variable_from_environ():
    yaml = "${VAR}"
    result = replace_variables_with_env(yaml, {"VAR": "value"})

    assert result == "value"


def test_replace_variable_from_default():
    yaml = "${VAR:default}"
    result = replace_variables_with_env(yaml, {})

    assert result == "default"


def test_replace_variable_environ_precedence():
    yaml = "${VAR:default}"
    result = replace_variables_with_env(yaml, {"VAR": "value"})

    assert result == "value"


def test_replace_multiple_variables():
    yaml = "${VAR1} ${VAR2}"
    result = replace_variables_with_env(yaml, {"VAR1": "1", "VAR2": "2"})

    assert result == "1 2"


def test_replace_variables_recursive():
    settings = {
        "prop": "${VAR1}",
        "list": ["${VAR2}"],
        "dict": {
            "var": "${VAR3}",
            "subdict": {"var": "${VAR4}"},
        },
    }

    environ = {"VAR1": "1", "VAR2": "2", "VAR3": "3", "VAR4": "4"}
    result = replace_variables_recursive(settings, environ)

    assert result == {
        "prop": "1",
        "list": ["2"],
        "dict": {
            "var": "3",
            "subdict": {"var": "4"},
        },
    }



import jinja2
import pytest
from pydantic import ValidationError

from selva.ext.templates.jinja.settings import JinjaTemplateSettings


class MyUndefined(jinja2.Undefined):
    pass


def finalize():
    pass


def autoescape():
    pass


class MyBytecodeCache(jinja2.BytecodeCache):
    pass


my_bytecode_cache = MyBytecodeCache()


def test_jinja_settings():
    settings = JinjaTemplateSettings.model_validate(
        {
            "undefined": f"{MyUndefined.__module__}:{MyUndefined.__qualname__}",
            "finalize": f"{finalize.__module__}:{finalize.__qualname__}",
            "autoescape": f"{autoescape.__module__}:{autoescape.__qualname__}",
            "bytecode_cache": f"{my_bytecode_cache.__module__}:my_bytecode_cache",
        },
    )

    assert settings.undefined is MyUndefined
    assert settings.finalize is finalize
    assert settings.autoescape is autoescape
    assert settings.bytecode_cache is my_bytecode_cache


@pytest.mark.parametrize(
    "value,expected",
    [
        (True, True),
        ("true", True),
        ("True", True),
        (False, False),
        ("false", False),
        ("False", False),
    ],
)
def test_autoescape_bool(value, expected):
    settings = JinjaTemplateSettings.model_validate({"autoescape": value})
    assert settings.autoescape is expected


def test_invalid_undefined_should_fail():
    with pytest.raises(ValidationError):
        JinjaTemplateSettings.model_validate({"undefined": 1})


def test_invalid_import_undefined_should_fail():
    with pytest.raises(ValidationError):
        JinjaTemplateSettings.model_validate({"undefined": "does.not.exist"})


def test_invalid_finalize_should_fail():
    with pytest.raises(ValidationError):
        JinjaTemplateSettings.model_validate({"finalize": 1})


def test_invalid_import_finalize_should_fail():
    with pytest.raises(ValidationError):
        JinjaTemplateSettings.model_validate({"finalize": "does.not.exist"})


def test_invalid_autoescape_should_fail():
    with pytest.raises(ValidationError):
        model = JinjaTemplateSettings.model_validate({"autoescape": "invalid"})
        repr(model)


def test_invalid_import_autoescape_should_fail():
    with pytest.raises(ValidationError):
        JinjaTemplateSettings.model_validate({"autoescape": "does.not.exist"})



import socket
from types import NoneType
from typing import Annotated, Any, Literal, Self

from pydantic import BaseModel, ConfigDict, Field, model_serializer, model_validator
from redis.backoff import (
    AbstractBackoff,
    ConstantBackoff,
    DecorrelatedJitterBackoff,
    EqualJitterBackoff,
    ExponentialBackoff,
    FullJitterBackoff,
    NoBackoff,
)
from redis.retry import Retry

from selva._util.pydantic import DottedPath


class NoBackoffSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")


class ConstantBackoffSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    backoff: int


class ExponentialBackoffSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    cap: int = None
    base: int = None


class BackoffSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    no_backoff: NoneType = None
    constant: ConstantBackoffSchema = None
    exponential: ExponentialBackoffSchema = None
    full_jitter: ExponentialBackoffSchema = None
    equal_jitter: ExponentialBackoffSchema = None
    decorrelated_jitter: ExponentialBackoffSchema = None

    @model_validator(mode="before")
    @classmethod
    def verify_backoff(cls, data):
        if len(data) != 1:
            raise ValueError("Only one backoff value can be set")

        return data


class RetrySchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    backoff: BackoffSchema
    retries: int
    supported_errors: tuple[DottedPath[type[Exception]], ...] = None

    @staticmethod
    def build_backoff(backoff: BackoffSchema) -> AbstractBackoff:
        if "no_backoff" in backoff.model_fields_set:
            return NoBackoff()

        if value := backoff.constant:
            return ConstantBackoff(**value.model_dump(exclude_unset=True))

        if value := backoff.exponential:
            return ExponentialBackoff(**value.model_dump(exclude_unset=True))

        if value := backoff.full_jitter:
            return FullJitterBackoff(**value.model_dump(exclude_unset=True))

        if value := backoff.equal_jitter:
            return EqualJitterBackoff(**value.model_dump(exclude_unset=True))

        if value := backoff.decorrelated_jitter:
            return DecorrelatedJitterBackoff(**value.model_dump(exclude_unset=True))

        raise ValueError("No value defined for 'backoff'")

    @model_serializer(when_used="unless-none")
    def serialize_model(self) -> dict[str, Any]:
        result = {
            "backoff": self.build_backoff(self.backoff),
            "retries": self.retries,
        }

        if supported_errors := self.supported_errors:
            result["supported_errors"] = supported_errors

        return result


class SocketKeepaliveOptions(BaseModel):
    tcp_keepidle: Annotated[int, Field(alias="TCP_KEEPIDLE")] = None
    tcp_keepcnt: Annotated[int, Field(alias="TCP_KEEPCNT")] = None
    tcp_keepintvl: Annotated[int, Field(alias="TCP_KEEPINTVL")] = None

    @model_serializer(when_used="unless-none")
    def serialize_model(self) -> dict[int, int]:
        result = {}
        if self.tcp_keepidle:
            result[socket.TCP_KEEPIDLE] = self.tcp_keepidle
        if self.tcp_keepcnt:
            result[socket.TCP_KEEPCNT] = self.tcp_keepcnt
        if self.tcp_keepintvl:
            result[socket.TCP_KEEPINTVL] = self.tcp_keepintvl
        return result


class RedisOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")

    socket_timeout: float = None
    socket_connect_timeout: float = None
    socket_keepalive: bool = None
    socket_keepalive_options: dict[
        Literal["TCP_KEEPIDLE", "TCP_KEEPCNT", "TCP_KEEPINTVL"], int
    ] = None
    unix_socket_path: str = None
    encoding: str = None
    encoding_errors: Literal["strict", "ignore", "replace"] = None
    decode_responses: bool = None
    retry_on_timeout: bool = None
    retry_on_error: list = None
    ssl: bool = None
    ssl_keyfile: str = None
    ssl_certfile: str = None
    ssl_cert_reqs: str = None
    ssl_ca_certs: str = None
    ssl_ca_data: str = None
    ssl_check_hostname: bool = None
    max_connections: int = None
    single_connection_client: bool = None
    health_check_interval: int = None
    client_name: str = None
    lib_name: str = None
    lib_version: str = None
    retry: RetrySchema = None
    auto_close_connection_pool: bool = None
    protocol: int = None

    @model_serializer(when_used="unless-none", mode="wrap")
    def serialize_model(self, nxt):
        result = nxt(self)
        if retry := result.get("retry"):
            result["retry"] = Retry(**retry)
        return result


class RedisSettings(BaseModel):
    model_config = ConfigDict(extra="forbid")

    url: str = None
    host: str = None
    port: int = None
    db: int = None
    username: str = None
    password: str = None
    options: RedisOptions = None

    @model_validator(mode="after")
    def verify_either_url_or_components(self) -> Self:
        if self.url and (self.host or self.port or (self.db is not None)):
            raise ValueError(
                "Either 'url' should be provided, or 'host', 'port' and 'db'"
            )

        return self

    @model_serializer(when_used="unless-none")
    def serializer(self):
        data = {
            field: getattr(self, field)
            for field in set(self.model_fields_set)
            if field != "options"
        }

        if self.options:
            data |= self.options.model_dump(exclude_unset=True)

        return data



import copy
from http import HTTPStatus
from pathlib import Path

import pytest
from httpx import ASGITransport, AsyncClient

from selva.configuration import Settings
from selva.configuration.defaults import default_settings
from selva.web.application import Selva
from selva.web.middleware.files import uploaded_files_middleware

MIDDLEWARE = [
    f"{uploaded_files_middleware.__module__}:{uploaded_files_middleware.__name__}"
]


@pytest.fixture(autouse=True)
def chdir_fixture(monkeypatch):
    monkeypatch.chdir(Path(__file__).parent)


async def test_uploaded_file():
    settings = Settings(
        default_settings
        | {
            "application": f"{__package__}.application",
            "middleware": copy.copy(MIDDLEWARE),
        }
    )
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/uploads/lorem-ipsum.txt")

    assert response.status_code == HTTPStatus.OK
    assert "text/plain" in response.headers["Content-Type"]
    assert response.text == "Lorem ipsum dolor sit amet."


async def test_uploaded_files_path():
    settings = Settings(
        default_settings
        | {
            "application": f"{__package__}.application",
            "middleware": copy.copy(MIDDLEWARE),
            "uploadedfiles": default_settings["uploadedfiles"]
            | {
                "path": "media",
            },
        }
    )
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/uploads/lorem-ipsum.txt")
    assert response.status_code == HTTPStatus.NOT_FOUND

    response = await client.get("http://localhost:8000/media/lorem-ipsum.txt")
    assert response.status_code == HTTPStatus.OK
    assert "text/plain" in response.headers["Content-Type"]
    assert response.text == "Lorem ipsum dolor sit amet."


async def test_uploaded_files_root():
    settings = Settings(
        default_settings
        | {
            "application": f"{__package__}.application",
            "middleware": copy.copy(MIDDLEWARE),
            "uploadedfiles": default_settings["uploadedfiles"]
            | {
                "root": Path(__file__).parent / "resources" / "media",
            },
        }
    )
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/uploads/data.json")
    assert response.status_code == HTTPStatus.OK
    assert "application/json" in response.headers["Content-Type"]
    assert response.text == '{"message": "lorem ipsum"}'



from selva.di.container import Container
from selva.di.decorator import service


@service
class Service:
    pass


async def test_define_service_not_registered(ioc: Container):
    instance = Service()
    ioc.define(Service, instance)

    service = await ioc.get(Service)

    assert service is instance


async def test_define_service_already_registered(ioc: Container):
    ioc.register(Service)
    instance = Service()
    ioc.define(Service, instance)

    service = await ioc.get(Service)

    assert service is instance



import os
from importlib.util import find_spec
from urllib.parse import urlparse

import pytest

from selva.configuration.settings import _get_settings_nocache
from selva.ext.data.redis.service import make_service

REDIS_URL = os.getenv("REDIS_URL")
PARSED_URL = urlparse(REDIS_URL) if REDIS_URL else None


pytestmark = [
    pytest.mark.skipif(REDIS_URL is None, reason="REDIS_URL not defined"),
    pytest.mark.skipif(find_spec("redis") is None, reason="redis not present"),
]


async def test_url_from_environment_variables(monkeypatch):
    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__URL", REDIS_URL)
    settings = _get_settings_nocache()

    async for redis in make_service("default")(settings):
        connection_kwargs = redis.connection_pool.connection_kwargs
        assert connection_kwargs["host"] == PARSED_URL.hostname
        assert connection_kwargs["port"] == PARSED_URL.port
        assert connection_kwargs["db"] == int(PARSED_URL.path.strip("/"))
        assert connection_kwargs.get("username") == PARSED_URL.username
        assert connection_kwargs.get("password") == PARSED_URL.password


async def test_url_username_password_from_environment_variables(monkeypatch):
    url = f"redis://{PARSED_URL.hostname}:{PARSED_URL.port}/{PARSED_URL.path}"
    username = PARSED_URL.username
    password = PARSED_URL.password

    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__URL", url)
    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__USERNAME", username)
    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__PASSWORD", password)
    settings = _get_settings_nocache()

    async for redis in make_service("default")(settings):
        connection_kwargs = redis.connection_pool.connection_kwargs
        assert connection_kwargs["host"] == PARSED_URL.hostname
        assert connection_kwargs["port"] == PARSED_URL.port
        assert connection_kwargs["db"] == int(PARSED_URL.path.strip("/"))
        assert connection_kwargs.get("username") == PARSED_URL.username
        assert connection_kwargs.get("password") == PARSED_URL.password


async def test_url_components_from_environment_variables(monkeypatch):
    host = PARSED_URL.hostname
    port = str(PARSED_URL.port)
    database = PARSED_URL.path.strip("/")
    username = PARSED_URL.username
    password = PARSED_URL.password

    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__HOST", host)
    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__PORT", port)
    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__DB", database)
    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__USERNAME", username)
    monkeypatch.setenv("SELVA__DATA__REDIS__DEFAULT__PASSWORD", password)

    settings = _get_settings_nocache()

    async for redis in make_service("default")(settings):
        connection_kwargs = redis.connection_pool.connection_kwargs
        assert connection_kwargs["host"] == PARSED_URL.hostname
        assert connection_kwargs["port"] == PARSED_URL.port
        assert connection_kwargs["db"] == int(PARSED_URL.path.strip("/"))
        assert connection_kwargs.get("username") == PARSED_URL.username
        assert connection_kwargs.get("password") == PARSED_URL.password



from collections.abc import Callable

from selva._util.maybe_async import maybe_async
from selva.di.container import Container
from selva.di.service.parse import get_dependencies


async def call_with_dependencies(di: Container, target: Callable):
    dependencies = {
        name: await di.get(dep.service, name=dep.name, optional=dep.optional)
        for name, dep in get_dependencies(target)
    }

    return await maybe_async(target, **dependencies)



from importlib.util import find_spec

from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.ext.data.sqlalchemy.service import (
    engine_dict_service,
    make_engine_service,
    sessionmaker_service,
)


def init_extension(container: Container, settings: Settings):
    if find_spec("sqlalchemy") is None:
        raise ModuleNotFoundError(
            "Missing 'sqlalchemy'. Install 'selva' with 'sqlalchemy' extra."
        )

    for name in settings.data.sqlalchemy.connections:
        container.register(make_engine_service(name))

    container.register(engine_dict_service)
    container.register(sessionmaker_service)



import pytest

from selva.web.routing.decorator import get
from selva.web.routing.exception import DuplicateRouteError
from selva.web.routing.router import Router


def test_duplicate_route_should_raise_error():
    @get
    async def route1(self, request):
        pass

    @get
    async def route2(self, request):
        pass

    router = Router()

    with pytest.raises(DuplicateRouteError):
        router.route(route1)
        router.route(route2)



from pathlib import Path

from httpx import ASGITransport, AsyncClient

from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.web.application import Selva

path = str(Path(__file__).parent.absolute())
settings = Settings(
    default_settings
    | {
        "application": f"{__package__}.application",
        "extensions": ["selva.ext.templates.jinja"],
        "templates": {"jinja": {"paths": [path]}},
    }
)


async def test_render():
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/render")

    assert response.status_code == 200
    assert response.text == "Jinja"
    assert response.headers["Content-Length"] == str(len("Jinja"))
    assert "text/html" in response.headers["Content-Type"]


async def test_stream():
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/stream")

    assert response.status_code == 200
    assert response.text == "Jinja"
    assert "Content-Length" not in response.headers


async def test_define_content_type():
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/define_content_type")

    assert response.status_code == 200
    assert "text/defined" in response.headers["Content-Type"]


async def test_override_content_type():
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/override_content_type")

    assert response.status_code == 200
    assert "text/overriden" in response.headers["Content-Type"]


async def test_content_type_from_response():
    app = Selva(settings)
    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/content_type_from_response")

    assert response.status_code == 200
    assert "text/from_response" in response.headers["Content-Type"]



from copy import deepcopy
from pathlib import Path

from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.ext.templates.mako.service import MakoTemplate


async def test_render_template():
    path = str(Path(__file__).parent.absolute())
    settings = Settings(
        default_settings | {"templates": {"mako": {"directories": [path]}}}
    )

    template = MakoTemplate(settings)
    template.initialize()
    result = await template.render("template.html", {"variable": "Mako"})
    assert result == "Mako"


async def test_render_str():
    settings = Settings(deepcopy(default_settings))
    template = MakoTemplate(settings)
    template.initialize()
    result = await template.render_str("${variable}", {"variable": "Mako"})
    assert result == "Mako"



from importlib.util import find_spec

from selva.configuration.settings import Settings
from selva.di.container import Container

from .service import make_service


def init_extension(container: Container, settings: Settings):
    if find_spec("redis") is None:
        raise ModuleNotFoundError(
            "Missing 'redis'. Install 'selva' with 'redis' extra."
        )

    for name in settings.data.redis:
        container.register(make_service(name))



from typing import Annotated

from selva.di.container import Container
from selva.di.decorator import service
from selva.di.inject import Inject


@service
class Dependency:
    pass


@service
class ServiceInitialize:
    initialized = False

    def initialize(self):
        self.initialized = True


@service
class ServiceAsyncInitialize:
    initialized = False

    async def initialize(self):
        self.initialized = True


@service
class Service:
    pass


@service
class ServiceFinalize:
    def finalize(self):
        print("finalize")


@service
class ServiceAsyncFinalize:
    async def finalize(self):
        print("async finalize", flush=True)


@service
def factory_finalize() -> Service:
    yield Service()
    print("factory finalize", flush=True)


@service
async def factory_async_finalize() -> Service:
    yield Service()
    print("factory async finalize", flush=True)


@service
class FinalizerOrder1:
    def initialize(self):
        print("initialize 1", flush=True)

    def finalize(self):
        print("finalize 1", flush=True)


@service
class FinalizerOrder2:
    dep: Annotated[FinalizerOrder1, Inject]

    def initialize(self):
        print("initialize 2", flush=True)

    def finalize(self):
        print("finalize 2", flush=True)


async def test_call_initialize(ioc: Container):
    ioc.register(ServiceInitialize)

    instance = await ioc.get(ServiceInitialize)
    assert instance.initialized


async def test_call_async_initialize(ioc: Container):
    ioc.register(Dependency)
    ioc.register(ServiceAsyncInitialize)

    instance = await ioc.get(ServiceAsyncInitialize)
    assert instance.initialized


async def test_call_finalize(ioc: Container, capsys):
    ioc.register(ServiceFinalize)

    await ioc.get(ServiceFinalize)
    await ioc.run_finalizers()

    assert capsys.readouterr().out == "finalize\n"


async def test_call_async_finalize(ioc: Container, capsys):
    ioc.register(ServiceAsyncFinalize)

    await ioc.get(ServiceAsyncFinalize)
    await ioc.run_finalizers()

    assert capsys.readouterr().out == "async finalize\n"


async def test_call_factory_finalize(ioc: Container, capsys):
    ioc.register(factory_finalize)

    await ioc.get(Service)
    await ioc.run_finalizers()

    assert capsys.readouterr().out == "factory finalize\n"


async def test_call_factory_async_finalize(ioc: Container, capsys):
    ioc.register(factory_async_finalize)

    await ioc.get(Service)
    await ioc.run_finalizers()

    assert capsys.readouterr().out == "factory async finalize\n"


async def test_finalizer_order(ioc: Container, capsys):
    ioc.register(FinalizerOrder1)
    ioc.register(FinalizerOrder2)

    await ioc.get(FinalizerOrder2)
    await ioc.run_finalizers()

    expected = "initialize 1\ninitialize 2\nfinalize 2\nfinalize 1\n"
    assert capsys.readouterr().out == expected



import structlog
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine

from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.di.decorator import service
from selva.ext.data.sqlalchemy.settings import (
    SqlAlchemyEngineSettings,
    SqlAlchemySettings,
)

logger = structlog.get_logger()


def make_engine_service(name: str):
    @service(name=name if name != "default" else None)
    async def engine_service(settings: Settings) -> AsyncEngine:
        sa_settings = SqlAlchemyEngineSettings.model_validate(
            settings.data.sqlalchemy.connections[name]
        )
        url = sa_settings.get_url()

        if sa_options := sa_settings.options:
            options = sa_options.model_dump(exclude_unset=True)
        else:
            options = {}

        engine = create_async_engine(url, **options)
        yield engine
        await engine.dispose()

    return engine_service


@service
async def engine_dict_service(
    settings: Settings, di: Container
) -> dict[str, AsyncEngine]:
    return {
        db: await di.get(AsyncEngine, name=db if db != "default" else None)
        for db in settings.data.sqlalchemy.connections
    }


@service
async def sessionmaker_service(
    settings: Settings, engines_map: dict[str, AsyncEngine]
) -> async_sessionmaker:
    sqlalchemy_settings = SqlAlchemySettings.model_validate(settings.data.sqlalchemy)

    args = []

    if session_options := sqlalchemy_settings.session.options:
        kwargs = session_options.model_dump(exclude_unset=True)
    else:
        kwargs = {}

    if binds := sqlalchemy_settings.session.binds:
        binds_config = {}
        for mapper, engine_name in binds.items():
            if engine := engines_map.get(engine_name):
                binds_config[mapper] = engine
            else:
                raise ValueError(f"No engine with name '{engine_name}'")

        kwargs["binds"] = binds_config
    else:
        engine = engines_map.get("default")
        if not engine:
            name, engine = next(iter(engines_map.items()))
            logger.warning("connection for sqlalchemy session", connection=name)
        args.append(engine)

    return async_sessionmaker(*args, **kwargs)



import copy

import pytest

from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.ext.error import ExtensionMissingInitFunctionError, ExtensionNotFoundError
from selva.web.application import Selva


async def test_non_existent_extension_should_fail():
    settings = Settings(
        copy.copy(default_settings)
        | {
            "application": f"{__package__}.application",
            "extensions": ["does.not.exist"],
        }
    )

    app = Selva(settings)

    with pytest.raises(ExtensionNotFoundError):
        await app._lifespan_startup()


async def test_extension_missing_init_function_should_fail():
    settings = Settings(
        copy.copy(default_settings)
        | {
            "application": f"{__package__}.application",
            "extensions": [f"{__package__}.extension"],
        }
    )

    app = Selva(settings)

    with pytest.raises(ExtensionMissingInitFunctionError):
        await app._lifespan_startup()



from http import HTTPMethod

import pytest

from selva.web.routing.decorator import (
    ATTRIBUTE_HANDLER,
    HandlerInfo,
    HandlerType,
    delete,
    get,
    patch,
    post,
    put,
    route,
    websocket,
)
from selva.web.routing.exception import (
    HandlerMissingRequestArgumentError,
    HandlerNotAsyncError,
    HandlerRequestTypeError,
    HandlerUntypedParametersError,
)


@pytest.mark.parametrize(
    "decorator,action_type",
    [
        (get, HandlerType.GET),
        (post, HandlerType.POST),
        (put, HandlerType.PUT),
        (patch, HandlerType.PATCH),
        (delete, HandlerType.DELETE),
        (websocket, HandlerType.WEBSOCKET),
    ],
    ids=["get", "post", "put", "patch", "delete", "websocket"],
)
def test_route_decorator_without_path(decorator, action_type):
    async def handler(req):
        pass

    decorator(handler)

    assert hasattr(handler, ATTRIBUTE_HANDLER)
    assert getattr(handler, ATTRIBUTE_HANDLER) == HandlerInfo(action_type, "")


@pytest.mark.parametrize(
    "decorator,action_type",
    [
        (get, HandlerType.GET),
        (post, HandlerType.POST),
        (put, HandlerType.PUT),
        (patch, HandlerType.PATCH),
        (delete, HandlerType.DELETE),
        (websocket, HandlerType.WEBSOCKET),
    ],
    ids=["get", "post", "put", "patch", "delete", "websocket"],
)
def test_route_decorator_with_path(decorator, action_type):
    async def handler(req):
        pass

    decorator("path")(handler)

    assert hasattr(handler, ATTRIBUTE_HANDLER)
    assert getattr(handler, ATTRIBUTE_HANDLER) == HandlerInfo(action_type, "path")


def test_handler_with_less_than_one_parameter_should_fail():
    async def func0():
        pass

    async def func1(req):
        pass

    with pytest.raises(HandlerMissingRequestArgumentError):
        route(func0, method=HTTPMethod.GET, path="")

    route(func1, method=HTTPMethod.GET, path="")


def test_handler_with_wrong_request_parameter_should_fail():
    async def handler(req: str):
        pass

    with pytest.raises(HandlerRequestTypeError):
        route(handler, method=HTTPMethod.GET, path="")


def test_handler_with_untyped_parameters_should_fail():
    async def handler(req, untyped):
        pass

    with pytest.raises(HandlerUntypedParametersError):
        route(handler, method=HTTPMethod.GET, path="")


def test_non_async_handler_should_fail():
    def handler(req):
        pass

    with pytest.raises(HandlerNotAsyncError):
        route(handler, method=HTTPMethod.GET, path="")



from importlib.util import find_spec

from selva.configuration.settings import Settings
from selva.di.container import Container
from selva.ext.templates.jinja.service import JinjaTemplate

__all__ = ("JinjaTemplate",)


async def init_extension(container: Container, _settings: Settings):
    if find_spec("jinja2") is None:
        raise ModuleNotFoundError(
            "Missing 'jinja2'. Install 'selva' with 'jinja' extra."
        )

    container.register(JinjaTemplate)



from typing import Annotated

import pytest

from selva.di.container import Container
from selva.di.decorator import service
from selva.di.error import DependencyLoopError
from selva.di.inject import Inject


@service
class Service1:
    service2: Annotated["Service2", Inject]


@service
class Service2:
    service1: Annotated[Service1, Inject]


@service
def factory1(service2: Service2) -> Service1:
    return Service1(service2)


@service
def factory2(service1: Service1) -> Service2:
    return Service2(service1)


async def test_service_class(ioc: Container):
    ioc.register(Service1)
    ioc.register(Service2)

    service1 = await ioc.get(Service1)
    service2 = await ioc.get(Service2)

    assert service1 is service2.service1
    assert service2 is service1.service2


async def test_service_factory(ioc: Container):
    ioc.register(factory1)
    ioc.register(factory2)

    with pytest.raises(DependencyLoopError):
        await ioc.get(Service1)



from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.web.application import Selva


async def test_extension():
    settings = Settings(
        default_settings
        | {
            "application": "tests.ext.register_extension.application",
            "extensions": ["tests.ext.register_extension.extension"],
        }
    )

    app = Selva(settings)

    await app._lifespan_startup()

    assert settings.tested



import os
from http import HTTPStatus

import pytest
from httpx import ASGITransport, AsyncClient

from selva.configuration.defaults import default_settings
from selva.configuration.settings import Settings
from selva.web.application import Selva

REDIS_URL = os.environ.get("REDIS_URL")


@pytest.mark.skipif(REDIS_URL is None, reason="REDIS_URL not set")
@pytest.mark.parametrize(
    "application,database",
    [
        ("application", "default"),
        ("application_named", "other"),
    ],
    ids=["default", "named"],
)
async def test_application(application: str, database: str):
    settings = Settings(
        default_settings
        | {
            "application": f"{__package__}.{application}",
            "extensions": ["selva.ext.data.redis"],
            "data": {"redis": {database: {"url": REDIS_URL}}},
        }
    )

    app = Selva(settings)

    await app._lifespan_startup()

    client = AsyncClient(transport=ASGITransport(app=app))
    response = await client.get("http://localhost:8000/")

    await app._lifespan_shutdown()

    assert response.status_code == HTTPStatus.OK



import decimal
from decimal import Decimal
from pathlib import PurePath

from selva.web.converter.decorator import register_converter
from selva.web.exception import HTTPBadRequestException


@register_converter(str, str)
class StrParamConverter:
    @staticmethod
    def convert(value: str, _original_type: type[str]) -> str:
        return value


@register_converter(str, int)
class IntParamConverter:
    @staticmethod
    def convert(value: str, _original_type: type[int]) -> int:
        try:
            return int(value)
        except ValueError as err:
            raise HTTPBadRequestException() from err


@register_converter(str, float)
class FloatParamConverter:
    @staticmethod
    def convert(value: str, _original_type: type[float]) -> float:
        try:
            return float(value)
        except ValueError as err:
            raise HTTPBadRequestException() from err


@register_converter(str, Decimal)
class DecimalParamConverter:
    @staticmethod
    def convert(value: str, _original_type: type[Decimal]) -> Decimal:
        try:
            return Decimal(value)
        except decimal.InvalidOperation as err:
            raise HTTPBadRequestException() from err


@register_converter(str, bool)
class BoolParamConverter:
    TRUE_VALUES = ["1", "true"]
    FALSE_VALUES = ["0", "false"]

    @staticmethod
    def convert(value: str, _original_type: type[bool]) -> bool:
        if value.lower() in BoolParamConverter.TRUE_VALUES:
            return True
        if value.lower() in BoolParamConverter.FALSE_VALUES:
            return False

        raise HTTPBadRequestException()


@register_converter(str, PurePath)
class PurePathParamConverter:
    @staticmethod
    def convert(value: str, _original_type: type[PurePath]) -> PurePath:
        return PurePath(value)



from abc import ABC
from collections.abc import Mapping

from selva.web.converter.decorator import (
    register_converter,
    register_from_request,
    register_param_extractor,
)

__all__ = (
    "register_converter",
    "register_from_request",
    "register_param_extractor",
    "Json",
    "Form",
)


class Json(ABC):
    pass


class Form(Mapping, ABC):
    pass



from collections.abc import Callable
from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field

from selva._util.pydantic import DottedPath


class MakoTemplateSettings(BaseModel):
    model_config = ConfigDict(extra="forbid")

    directories: Annotated[
        list[str], Field(default_factory=lambda: ["resources/templates"])
    ]
    module_directory: str = None
    filesystem_checks: bool = False
    collection_size: int = None
    format_exceptions: bool = None
    error_handler: DottedPath = None
    encoding_errors: Literal[
        "strict", "ignore", "replace", "xmlcharrefreplace", "htmlentityreplace"
    ] = None
    cache_enabled: bool = None
    cache_impl: str = None
    cache_args: DottedPath[dict] = None
    modulename_callable: DottedPath[Callable] = None
    module_writer: DottedPath[Callable] = None
    default_filters: list[str] = None
    buffer_filters: list[str] = None
    strict_undefined: bool = None
    imports: list[str] = None
    future_imports: list[str] = None
    enable_loop: bool = None
    input_encoding: str = "utf-8"
    preprocessor: DottedPath[Callable] = None
    lexer_cls: DottedPath[type] = None
    include_error_handler: DottedPath[Callable] = None



from typing import Any

from selva.di.container import Container
from selva.di.decorator import service


@service
class MyInterceptor:
    async def intercept(self, instance: Any, _service_type: type):
        setattr(instance, "intercepted", True)


@service
class MyService:
    pass


async def test_intercept(ioc: Container):
    ioc.register(MyService)
    ioc.interceptor(MyInterceptor)

    instance = await ioc.get(MyService)

    assert getattr(instance, "intercepted")



import uuid

import structlog
from asgikit.requests import Request

from selva.configuration.settings import Settings
from selva.di.container import Container


async def request_id_middleware(app, settings: Settings, di: Container):
    async def handler(scope, receive, send):
        request = Request(scope, receive, send)

        request_id = request.headers.get("x-request-id", str(uuid.uuid4()))
        request["request_id"] = request_id
        structlog.contextvars.bind_contextvars(request_id=request_id)

        await app(scope, receive, send)

        structlog.contextvars.unbind_contextvars("request_id")

    return handler



from redis.asyncio import Redis

from selva.configuration.settings import Settings
from selva.di.decorator import service

from .settings import RedisSettings


def make_service(name: str):
    @service(name=name if name != "default" else None)
    async def redis_service(settings: Settings) -> Redis:
        redis_settings = RedisSettings.model_validate(dict(settings.data.redis[name]))

        kwargs = redis_settings.model_dump(exclude_unset=True)

        if url := kwargs.pop("url", ""):
            redis = Redis.from_url(url, **kwargs)
        else:
            redis = Redis(**kwargs)

        await redis.initialize()
        yield redis
        await redis.aclose()

    return redis_service



import itertools

import pytest
from redis.backoff import (
    ConstantBackoff,
    DecorrelatedJitterBackoff,
    EqualJitterBackoff,
    ExponentialBackoff,
    FullJitterBackoff,
    NoBackoff,
)

from selva.ext.data.redis.settings import (
    BackoffSchema,
    RedisOptions,
    RedisSettings,
    RetrySchema,
)


@pytest.mark.parametrize(
    "values",
    [
        {"host": "host"},
        {"port": 5432},
        {"db": 0},
        {"host": "host", "port": 5423},
        {"host": "host", "port": 5432, "db": 0},
    ],
    ids=[
        "host",
        "port",
        "db",
        "host_port",
        "host_port_db",
    ],
)
def test_mutually_exclusive_connection_properties(values: dict):
    with pytest.raises(ValueError):
        RedisSettings.model_validate({"url": "url"} | values)


@pytest.mark.parametrize(
    "backoff, cls",
    [
        ({"no_backoff": None}, NoBackoff),
        ({"constant": {"backoff": 1}}, ConstantBackoff),
        ({"exponential": {"cap": 1, "base": 1}}, ExponentialBackoff),
        ({"full_jitter": {"cap": 1, "base": 1}}, FullJitterBackoff),
        ({"equal_jitter": {"cap": 1, "base": 1}}, EqualJitterBackoff),
        ({"decorrelated_jitter": {"cap": 1, "base": 1}}, DecorrelatedJitterBackoff),
    ],
)
def test_retry_backoff(backoff, cls):
    model = RetrySchema.model_validate(
        {
            "backoff": backoff,
            "retries": 1,
        }
    )

    config = model.model_dump(exclude_none=True)
    assert isinstance(config["backoff"], cls)


@pytest.mark.parametrize(
    "values",
    [
        left | right
        for left, right in itertools.combinations(
            [
                {"no_backoff": None},
                {"constant": {"backoff": 1}},
                {"exponential": {"cap": 1, "base": 1}},
                {"full_jitter": {"cap": 1, "base": 1}},
                {"equal_jitter": {"cap": 1, "base": 1}},
                {"decorrelated_jitter": {"cap": 1, "base": 1}},
            ],
            2,
        )
    ],
)
def test_mutually_exclusive_retry_properties(values: dict):
    with pytest.raises(ValueError):
        BackoffSchema.model_validate(values)


def test_invalid_encode_errors_property():
    with pytest.raises(ValueError):
        RedisOptions.model_validate({"encoding_errors": "invalid"})
